#include "keyboard.h"
#include "pd8049.h"
#include "schedule.h"
#include "tape.h"


// ٥ID
#define	EID_AKEY	(1)

static const struct {
	int Key;			// SDLΥ
	Uint8 Keys[8];		// P6Υ
} P6Keys[] = {
		//				  STD	STD+S	GRAPH		+S		+S	CTRL
		//				--------------------------------------------------------------
		//
	{ SDLK_UNKNOWN,		{ 0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//				  BS			????	BS		BS		BS		BS
	{ SDLK_BACKSPACE,	{ 0x08,	0x00,	0x7f,	0x08,	0x08,	0x08,	0x08,	0x00 } },
		//				  TAB	TAB				TAB		TAB		TAB		TAB
	{ SDLK_TAB,			{ 0x09,	0x09,	0x00,	0x09,	0x09,	0x09,	0x09,	0x00 } },
		//				  RET	RET				RET		RET		RET		RET
	{ SDLK_RETURN,		{ 0x0d,	0x0d,	0x00,	0x0d,	0x0d,	0x0d,	0x0d,	0x00 } },
		//								
	{ SDLK_PAUSE,		{ 0x00,	0x00,	0xfe,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//				  ESC	ESC				ESC		ESC		ESC		ESC
	{ SDLK_ESCAPE,		{ 0x1b,	0x1b,	0x00,	0x1b,	0x1b,	0x1b,	0x1b,	0x00 } },
		//				  Space	Space			Space	Space	Space	Space
	{ SDLK_SPACE,		{ 0x20,	0x20,	0x00,	0x20,	0x20,	0x20,	0x20,	0x00 } },
		//				  	<										
	{ SDLK_COMMA,		{ 0x2c,	0x3c,	0x1f,	0xe8,	0xa4,	0xc8,	0xa4,	0x00 } },
		//				  -		=								
	{ SDLK_MINUS,		{ 0x2d,	0x3d,	0x17,	0xee,	0x00,	0xce,	0x00,	0x00 } },
		//				  .		>										
	{ SDLK_PERIOD,		{ 0x2e,	0x3e,	0x1d,	0xf9,	0xa1,	0xd9,	0xa1,	0x00 } },
		//				  /		?		ڡ 						
	{ SDLK_SLASH,		{ 0x2f,	0x3f,	0x80,	0xf2,	0xa5,	0xd2,	0xa5,	0x00 } },
		//				  0												
	{ SDLK_0,			{ 0x30,	0x00,	0x0f,	0xfc,	0x86,	0xdc,	0xa6,	0x00 } },
		//				  1		!								
	{ SDLK_1,			{ 0x31,	0x21,	0x07,	0xe7,	0x00,	0xc7,	0x00,	0x00 } },
		//				  2		"								
	{ SDLK_2,			{ 0x32,	0x22,	0x01,	0xec,	0x00,	0xcc,	0x00,	0x00 } },
		//				  3		#										
	{ SDLK_3,			{ 0x33,	0x23,	0x02,	0x91,	0x87,	0xb1,	0xa7,	0x00 } },
		//				  4		$										
	{ SDLK_4,			{ 0x34,	0x24,	0x03,	0x93,	0x89,	0xb3,	0xa9,	0x00 } },
		//				  5		%										
	{ SDLK_5,			{ 0x35,	0x25,	0x04,	0x94,	0x8a,	0xb4,	0xaa,	0x00 } },
		//				  6		&										
	{ SDLK_6,			{ 0x36,	0x26,	0x05,	0x95,	0x8b,	0xb5,	0xab,	0x00 } },
		//				  7		'										
	{ SDLK_7,			{ 0x37,	0x27,	0x06,	0xf4,	0x8c,	0xd4,	0xac,	0x00 } },
		//				  8		(		ɴ								
	{ SDLK_8,			{ 0x38,	0x28,	0x0d,	0xf5,	0x8d,	0xd5,	0xad,	0x00 } },
		//				  9		)										
	{ SDLK_9,			{ 0x39,	0x29,	0x0e,	0xf6,	0x8e,	0xd6,	0xae,	0x00 } },
		//				  :		*		ϡ					
	{ SDLK_COLON,		{ 0x3a,	0x2a,	0x81,	0x99,	0x00,	0xb9,	0x00,	0x00 } },
		//				  ;		+							
	{ SDLK_SEMICOLON,	{ 0x3b,	0x2b,	0x82,	0xfa,	0x00,	0xda,	0x00,	0x00 } },
		//				  @										
	{ SDLK_AT,			{ 0x40,	0x00,	0x00,	0xde,	0x00,	0xde,	0x00,	0x00 } },
		//				  [												
	{ SDLK_LEFTBRACKET,	{ 0x5b,	0x00,	0x84,	0xdf,	0xa2,	0xdf,	0xa2,	0x00 } },
		//				  									
	{ SDLK_BACKSLASH,	{ 0x5c,	0x00,	0x09,	0xb0,	0x00,	0xb0,	0x00,	0x00 } },
		//				  ]												
	{ SDLK_RIGHTBRACKET,{ 0x5d,	0x00,	0x85,	0xf1,	0xa3,	0xd1,	0xa3,	0x00 } },
		//				  ^										
	{ SDLK_CARET,		{ 0x5e,	0x00,	0x00,	0xed,	0x00,	0xcd,	0x00,	0x00 } },
		//						_							
	{ SDLK_UNDERSCORE,	{ 0x00,	0x5f,	0x83,	0xfb,	0x00,	0xdb,	0x00,	0x00 } },
		//				  a		A												A
	{ SDLK_a,			{ 0x61,	0x41,	0x00,	0xe1,	0x00,	0xc1,	0x00,	0x01 } },
		//				  b		B												B
	{ SDLK_b,			{ 0x62,	0x42,	0x1b,	0x9a,	0x00,	0xba,	0x00,	0x02 } },
		//				  c		C												C
	{ SDLK_c,			{ 0x63,	0x43,	0x1a,	0x9f,	0x00,	0xbf,	0x00,	0x03 } },
		//				  d		D												D
	{ SDLK_d,			{ 0x64,	0x44,	0x14,	0x9c,	0x00,	0xbc,	0x00,	0x04 } },
		//				  e		E												E
	{ SDLK_e,			{ 0x65,	0x45,	0x18,	0x92,	0x88,	0xb2,	0xa8,	0x05 } },
		//				  f		F												F
	{ SDLK_f,			{ 0x66,	0x46,	0x15,	0xea,	0x00,	0xca,	0x00,	0x06 } },
		//				  g		G												G
	{ SDLK_g,			{ 0x67,	0x47,	0x13,	0x97,	0x00,	0xb7,	0x00,	0x07 } },
		//				  h		H												H
	{ SDLK_h,			{ 0x68,	0x48,	0x0a,	0x98,	0x00,	0xb8,	0x00,	0x08 } },
		//				  i		I												I
	{ SDLK_i,			{ 0x69,	0x49,	0x16,	0xe6,	0x00,	0xc6,	0x00,	0x09 } },
		//				  j		J												J
	{ SDLK_j,			{ 0x6a,	0x4a,	0x00,	0xef,	0x00,	0xcf,	0x00,	0x0a } },
		//				  k		K												K
	{ SDLK_k,			{ 0x6b,	0x4b,	0x00,	0xe9,	0x00,	0xc9,	0x00,	0x0b } },
		//				  l		L												L
	{ SDLK_l,			{ 0x6c,	0x4c,	0x1e,	0xf8,	0x00,	0xd8,	0x00,	0x0c } },
		//				  m		M		ʬ										M
	{ SDLK_m,			{ 0x6d,	0x4d,	0x0b,	0xf3,	0x00,	0xd3,	0x00,	0x0d } },
		//				  n		N												N
	{ SDLK_n,			{ 0x6e,	0x4e,	0x00,	0xf0,	0x00,	0xd0,	0x00,	0x0e } },
		//				  o		O												O
	{ SDLK_o,			{ 0x6f,	0x4f,	0x00,	0xf7,	0x00,	0xd7,	0x00,	0x0f } },
		//				  p		P												P
	{ SDLK_p,			{ 0x70,	0x50,	0x10,	0x9e,	0x00,	0xbe,	0x00,	0x10 } },
		//				  q		Q												Q
	{ SDLK_q,			{ 0x71,	0x51,	0x00,	0xe0,	0x00,	0xc0,	0x00,	0x11 } },
		//				  r		R												R
	{ SDLK_r,			{ 0x72,	0x52,	0x12,	0x9d,	0x00,	0xbd,	0x00,	0x12 } },
		//				  s		S												S
	{ SDLK_s,			{ 0x73,	0x53,	0x0c,	0xe4,	0x00,	0xc4,	0x00,	0x13 } },
		//				  t		T												T
	{ SDLK_t,			{ 0x74,	0x54,	0x19,	0x96,	0x00,	0xb6,	0x00,	0x14 } },
		//				  u		U												U
	{ SDLK_u,			{ 0x75,	0x55,	0x00,	0xe5,	0x00,	0xc5,	0x00,	0x15 } },
		//				  v		V												V
	{ SDLK_v,			{ 0x76,	0x56,	0x11,	0xeb,	0x00,	0xcb,	0x00,	0x16 } },
		//				  w		W												W
	{ SDLK_w,			{ 0x77,	0x57,	0x00,	0xe3,	0x00,	0xc3,	0x00,	0x17 } },
		//				  x		X												X
	{ SDLK_x,			{ 0x78,	0x58,	0x1c,	0x9b,	0x00,	0xbb,	0x00,	0x18 } },
		//				  y		Y		ǯ										Y
	{ SDLK_y,			{ 0x79,	0x59,	0x08,	0xfd,	0x00,	0xdd,	0x00,	0x19 } },
		//				  z		Z												Z
	{ SDLK_z,			{ 0x7a,	0x5a,	0x00,	0xe2,	0x8f,	0xc2,	0xaf,	0x1a } },
		//				  DEL	DEL		????	DEL		DEL		DEL		DEL
	{ SDLK_DELETE,		{ 0x08,	0x08,	0x7f,	0x08,	0x08,	0x08,	0x08,	0x00 } },
		//				  [0]	[0]		[0]		[0]		[0]		[0]		[0]
	{ SDLK_KP0,			{ 0x30,	0x30,	0x30,	0x30,	0x30,	0x30,	0x30,	0x00 } },
		//				  [1]	[1]		[1]		[1]		[1]		[1]		[1]
	{ SDLK_KP1,			{ 0x31,	0x31,	0x31,	0x31,	0x31,	0x31,	0x31,	0x00 } },
		//				  [2]	[2]		[2]		[2]		[2]		[2]		[2]
	{ SDLK_KP2,			{ 0x32,	0x32,	0x32,	0x32,	0x32,	0x32,	0x32,	0x00 } },
		//				  [3]	[3]		[3]		[3]		[3]		[3]		[3]
	{ SDLK_KP3,			{ 0x33,	0x33,	0x33,	0x33,	0x33,	0x33,	0x33,	0x00 } },
		//				  [4]	[4]		[4]		[4]		[4]		[4]		[4]
	{ SDLK_KP4,			{ 0x34,	0x34,	0x34,	0x34,	0x34,	0x34,	0x34,	0x00 } },
		//				  [5]	[5]		[5]		[5]		[5]		[5]		[5]
	{ SDLK_KP5,			{ 0x35,	0x35,	0x35,	0x35,	0x35,	0x35,	0x35,	0x00 } },
		//				  [6]	[6]		[6]		[6]		[6]		[6]		[6]
	{ SDLK_KP6,			{ 0x36,	0x36,	0x36,	0x36,	0x36,	0x36,	0x36,	0x00 } },
		//				  [7]	[7]		[7]		[7]		[7]		[7]		[7]
	{ SDLK_KP7,			{ 0x37,	0x37,	0x37,	0x37,	0x37,	0x37,	0x37,	0x00 } },
		//				  [8]	[8]		[8]		[8]		[8]		[8]		[8]
	{ SDLK_KP8,			{ 0x38,	0x38,	0x38,	0x38,	0x38,	0x38,	0x38,	0x00 } },
		//				  [9]	[9]		[9]		[9]		[9]		[9]		[9]
	{ SDLK_KP9,			{ 0x39,	0x39,	0x39,	0x39,	0x39,	0x39,	0x39,	0x00 } },
		//				  [.]	[.]		[.]		[.]		[.]		[.]		[.]
	{ SDLK_KP_PERIOD,	{ 0x2e,	0x2e,	0x2e,	0x2e,	0x2e,	0x2e,	0x2e,	0x00 } },
		//				  [/]	[/]		[/]		[/]		[/]		[/]		[/]
	{ SDLK_KP_DIVIDE,	{ 0x2f,	0x2f,	0x2f,	0x2f,	0x2f,	0x2f,	0x2f,	0x00 } },
		//				  [*]	[*]		[*]		[*]		[*]		[*]		[*]
	{ SDLK_KP_MULTIPLY,	{ 0x2a,	0x2a,	0x2a,	0x2a,	0x2a,	0x2a,	0x2a,	0x00 } },
		//				  [-]	[-]		[-]		[-]		[-]		[-]		[-]
	{ SDLK_KP_MINUS,	{ 0x2d,	0x2d,	0x2d,	0x2d,	0x2d,	0x2d,	0x2d,	0x00 } },
		//				  [+]	[+]		[+]		[+]		[+]		[+]		[+]
	{ SDLK_KP_PLUS,		{ 0x2b,	0x2b,	0x2b,	0x2b,	0x2b,	0x2b,	0x2b,	0x00 } },
		//				  [RET]	[RET]	[RET]	[RET]	[RET]	[RET]	[RET]
	{ SDLK_KP_ENTER,	{ 0x0d,	0x0d,	0x0d,	0x0d,	0x0d,	0x0d,	0x0d,	0x00 } },
		//				  											
	{ SDLK_UP,			{ 0x1e,	0x1e,	0x1e,	0x1e,	0x1e,	0x1e,	0x1e,	0x00 } },
		//				  											
	{ SDLK_DOWN,		{ 0x1f,	0x1f,	0x1f,	0x1f,	0x1f,	0x1f,	0x1f,	0x00 } },
		//				  											
	{ SDLK_RIGHT,		{ 0x1c,	0x1c,	0x1c,	0x1c,	0x1c,	0x1c,	0x1c,	0x00 } },
		//				  											
	{ SDLK_LEFT,		{ 0x1d,	0x1d,	0x1d,	0x1d,	0x1d,	0x1d,	0x1d,	0x00 } },
		//				  INS	INS				INS		INS		INS		INS
	{ SDLK_INSERT,		{ 0x12,	0x12,	0x12,	0x12,	0x12,	0x12,	0x12,	0x00 } },
		//				  HOME	HOME			HOME	HOME	HOME	HOME
	{ SDLK_HOME,		{ 0x0c,	0x0b,	0x0c,	0x0c,	0x0b,	0x0c,	0x0b,	0x00 } },
		//				  STOP	STOP	STOP	STOP	STOP	STOP	STOP
	{ SDLK_END,			{ 0xfa,	0xfa,	0xfa,	0xfa,	0xfa,	0xfa,	0xfa,	0x00 } },
		//				  PAGE	PAGE	PAGE	PAGE	PAGE	PAGE	PAGE
	{ SDLK_PAGEUP,		{ 0xfe,	0xfe,	0xfc,	0xfe,	0xfe,	0xfe,	0xfe,	0x00 } },
		//								MODE
	{ SDLK_PAGEDOWN,	{ 0x00,	0x00,	0xfd,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//						F6		F1
	{ SDLK_F1,			{ 0x00,	0xf5,	0xf0,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//						F7		F2
	{ SDLK_F2,			{ 0x00,	0xf6,	0xf1,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//						F8		F3
	{ SDLK_F3,			{ 0x00,	0xf7,	0xf2,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//						F9		F4
	{ SDLK_F4,			{ 0x00,	0xf8,	0xf3,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//						F10		F5
	{ SDLK_F5,			{ 0x00,	0xf9,	0xf4,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//
	{ SDLK_F6,			{ 0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//
	{ SDLK_F7,			{ 0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//
	{ SDLK_F8,			{ 0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//
	{ SDLK_F9,			{ 0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//
	{ SDLK_F10,			{ 0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//
	{ SDLK_CAPSLOCK,	{ 0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//
	{ SDLK_RCTRL,		{ 0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//
	{ SDLK_LCTRL,		{ 0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//
	{ SDLK_LSHIFT,		{ 0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//
	{ SDLK_RSHIFT,		{ 0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//
	{ SDLK_LALT,		{ 0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//
	{ SDLK_RALT,		{ 0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//
	{ SDLK_SYSREQ,		{ 0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//								CAPS
	{ SDLK_SCROLLOCK,	{ 0x00,	0x00,	0xfb,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//
	{ SDLK_F11,			{ 0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00 } },
		//
	{ SDLK_F12,			{ 0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00,	0x00 } }
};

enum KeyGroup
{
	KEYSTD = 0,
	KEYSHIFT,
	KEYGRAPH,
	KEYKANA,
	KEYKANASHIFT,
	KEYKKANA,
	KEYKKANASHIFT,
	KEYCTRL
};



////////////////////////////////////////////////////////////////
// 󥹥ȥ饯
////////////////////////////////////////////////////////////////
KEY6::KEY6( VM6 *vm ) : P6DEVENT(vm)
{
	ON_SHIFT =			// SHIFT
	ON_GRAPH =			// GRAPH
	ON_KANA  =			// 
	ON_KKANA =			// 
	ON_CTRL  =			// CTRL
	ON_CAPS  = FALSE;	// CAPS
	
	// ơ֥
	for( int i=0; i < SDLK_LAST; i++ ) P6KeyTable[i] = (Uint8 *)P6Keys[SDLK_UNKNOWN].Keys;
	for( int i=0; i < COUNTOF(P6Keys); i++ ) P6KeyTable[P6Keys[i].Key] = (Uint8 *)P6Keys[i].Keys;
	
	// ư̵
	ZeroMemory( &ak, sizeof( AKEY ) );
}

KEY60::KEY60( VM6 *vm ) : KEY6(vm)
{
	// MODE  CAPS ̵
	P6KeyTable[SDLK_PAGEDOWN]  = (Uint8 *)P6Keys[SDLK_UNKNOWN].Keys;
	P6KeyTable[SDLK_SCROLLOCK] = (Uint8 *)P6Keys[SDLK_UNKNOWN].Keys;
}


////////////////////////////////////////////////////////////////
// ǥȥ饯
////////////////////////////////////////////////////////////////
KEY6::~KEY6( void )
{
	if( ak.Buffer ) delete [] ak.Buffer;
}

KEY60::~KEY60( void ){}


////////////////////////////////////////////////////////////////
// ٥ȥХåؿ
//
// :	id		٥ID
//			clock	å
// :	ʤ
////////////////////////////////////////////////////////////////
void KEY6::EventCallback( int id, int clock )
{
	switch( id ){
	case EID_AKEY:	// ưϽ
		if( !vm->cmt->IsRelay() ){	// TAPEäƤϥߤȯʤ
			BYTE key = GetAutoKey();
			if( key ) vm->subcpu->ReqKeyIntr( 0, key );
		}
		break;
	}
}


////////////////////////////////////////////////////////////////
// ư1ʸ
//   10msȤ˸ƤФ
//
// :	ʤ
// :	BYTE	P6Υ
////////////////////////////////////////////////////////////////
char KEY6::GetAutoKey( void )
{
	if( ak.Num > 0 ){
		// 졼OFFԤ
		if( ak.Relay ){
			if( vm->cmt->IsRelay() ) return 0;
			else                     ak.Relay = FALSE;
		}
		
		// Ԥ?
		if( ak.Wait > 0 ){
			ak.Wait--;
			return 0;
		}
		
		// ʸ
		BYTE dat = ak.Buffer[ak.Seek++];
		switch( dat ){
		case 0x0a:	// '\r' 졼OFFԤ
			ak.Relay = TRUE;
			dat = 0x0d;
		case 0x0d:	// '\n' ?
			ak.Wait = 150;	// Ԥ1.5sec
			break;
		default:	// ̤ʸ
			ak.Wait = 2;	// Ԥ20msec
		}
		// Хåեʤ齪λ
		if( --ak.Num <= 0 ){
			vm->sche->Del( this, EID_AKEY );
			delete [] ak.Buffer;
			ak.Buffer = NULL;
		}
		return dat;
	}
	return 0;
}


////////////////////////////////////////////////////////////////
// ưʸ
//
// :	str		ʸ
//			num		ʸĹ
// :	BOOL	TRUE: FALSE:
////////////////////////////////////////////////////////////////
BOOL KEY6::SetAutoKey( char *str, int num )
{
	// Хåեö
	if( ak.Buffer ) delete [] ak.Buffer;
	
	// Хåե
	ak.Buffer = new char[num+1];
	if( !num ) return FALSE;
	
	// ʸ򥳥ԡ
	strncpy( ak.Buffer, str, num );
	
	ak.Num    = num;		// Ĥʸ
	ak.Wait   = 200;		// Ԥ(2sec)
	ak.Relay  = FALSE;		// 졼åOFFԤե饰
	ak.Seek   = 0;			// ɹߥݥ
	
	return vm->sche->Add( this, EID_AKEY, 10, EV_LOOP|EV_MS );
}


////////////////////////////////////////////////////////////////
// 
//
// :	repeat	ԡȤδֳ(ms) 0̵
// :	BOOL	TRUE: FALSE:
////////////////////////////////////////////////////////////////
BOOL KEY6::Init( int repeat )
{
	PRINTD( KEY_LOG, "[KEY][Init]\n" );
	
	// ü쥭 ե饰
	ON_SHIFT =
	ON_GRAPH =
	ON_KANA  =
	ON_KKANA =
	ON_CTRL  =
	ON_STOP  =
	ON_CAPS  = FALSE;
	
	// ԡ
	if( repeat ) SDL_EnableKeyRepeat( 500, repeat );
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// 
//
// :	event	SDLΥϥ٥
// :	ʤ
////////////////////////////////////////////////////////////////
void KEY6::ScanKeyboard( SDL_Event *event )
{
	PRINTD( KEY_LOG, "[KEY][ScanKeyboard]\n" );
	
	BOOL ON_FUNC  = FALSE;	// ե󥯥󥭡Ȥե饰
	Uint8 KeyData = 0;		// P6Υ
	
	int KeyCode  = event->key.keysym.sym;							// SDLΥ
	BOOL KeyPUSH = ( event->type == SDL_KEYDOWN ) ? TRUE : FALSE;	// ե饰
	
	// ڻܸ֡쥭ܡб(ATߴ)
	switch( event->key.keysym.scancode ){	// ϡɥͭΥ󥳡
	case 0x2b:	KeyCode = SDLK_UNDERSCORE;	break;	// _
	case 0x7d:	KeyCode = SDLK_BACKSLASH;	break;	// 
	case 0x90:	KeyCode = SDLK_CARET;		break;	// ^
	case 0x91:	KeyCode = SDLK_AT;			break;	// @
	case 0x92:	KeyCode = SDLK_COLON;		break;	// :
	}
	
	// ü쥭Ƚ(ON-OFF֤Ƚ)
	switch( KeyCode ){	// 
	case SDLK_LCTRL:	// CTRL
	case SDLK_RCTRL:	// CTRL
		ON_CTRL = KeyPUSH;
		break;
		
	case SDLK_LSHIFT:	// SHIFT
	case SDLK_RSHIFT:	// SHIFT
		ON_SHIFT = KeyPUSH;
		break;
		
	case SDLK_LALT:		// ALT
	case SDLK_RALT:		// ALT
		ON_GRAPH = KeyPUSH;
		break;
		
	// ̥Ƚ
	default:
		if( KeyPUSH ){	// ?
			switch( KeyCode ){	// 
			case SDLK_SCROLLOCK:	// CAPS
				// CAPSͭ?
				if( P6KeyTable[SDLK_SCROLLOCK][2] ){
					ON_CAPS = ON_CAPS ? FALSE : TRUE;
					ON_FUNC = TRUE;
				}
				break;
				
			case SDLK_PAGEDOWN:		// MODE
				ON_FUNC = TRUE;
				break;
				
			case SDLK_PAUSE:		// 
				ON_KANA = ON_KANA ? FALSE : TRUE;
				ON_FUNC = TRUE;
				break;
				
			case SDLK_END:			// STOP
				ON_STOP = TRUE;
				break;
				
			case SDLK_PAGEUP:		// PAGE or /ڤؤ
				if( ON_SHIFT ){
					ON_KKANA = ON_KKANA ? FALSE : TRUE;
					ON_FUNC  = TRUE;
				}
				break;
				
			case SDLK_F1:			// F01F05ʥե󥯥󥭡
			case SDLK_F2:
			case SDLK_F3:
			case SDLK_F4:
			case SDLK_F5:
				ON_FUNC = TRUE;
				break;
			}
			
			// SDLΥɤP6ΥɤѴ
			if( ON_FUNC ){
				if( KeyCode < SDLK_F1 || KeyCode > SDLK_F5 )
					KeyData = P6KeyTable[KeyCode][KEYGRAPH];
				else	// F06F10 ν
					if( ON_SHIFT ) KeyData = P6KeyTable[KeyCode][KEYSHIFT];
					else           KeyData = P6KeyTable[KeyCode][KEYGRAPH];
			}else{
				if( ON_KANA ){
					if( ON_KKANA )
						if( ON_SHIFT ) KeyData = P6KeyTable[KeyCode][KEYKKANASHIFT];
						else           KeyData = P6KeyTable[KeyCode][KEYKKANA];
					else
						if( ON_SHIFT ) KeyData = P6KeyTable[KeyCode][KEYKANASHIFT];
						else           KeyData = P6KeyTable[KeyCode][KEYKANA];
				}else{
					if( ON_GRAPH )
						KeyData = P6KeyTable[KeyCode][KEYGRAPH];
					else
						if( ON_CTRL )
							KeyData = P6KeyTable[KeyCode][KEYCTRL];
						else
							if( ON_SHIFT ) KeyData = P6KeyTable[KeyCode][KEYSHIFT];
							else           KeyData = P6KeyTable[KeyCode][KEYSTD];
				}
			}
			
			// ߽
			if( KeyData ){	// ͭʥ?
				// flag
				// bit 0 : ON_STOP
				// bit 1 : ON_GRAPH
				// bit 2 : ON_FUNC
				
				vm->subcpu->ReqKeyIntr( ON_STOP ? 1 : 0 | ON_GRAPH ? 2 : 0 | ON_FUNC ? 4 : 0, KeyData );
				ON_STOP = FALSE;
			}
		}
	}
	
	PRINTD2( KEY_LOG, " KeyData:%02X KeyCode:%02X ", KeyData, KeyCode );
	PRINTD1( KEY_LOG, " %s", ON_STOP  ? "(STOP)" : "" );
	PRINTD1( KEY_LOG, " %s", ON_GRAPH ? "(GRAPH)" : "" );
	PRINTD1( KEY_LOG, " %s", ON_FUNC  ? "(FUNC)\n" : "\n" );
}


////////////////////////////////////////////////////////////////
// 륭ּ
//
// :	ʤ
// :	BYTE	륭ξ 1:ON 0:OFF
//				bit7: ڡ
//				bit6: 0
//				bit5: 
//				bit4: 
//				bit3: 
//				bit2: 
//				bit1: STOP
//				bit0: SHIFT
////////////////////////////////////////////////////////////////
BYTE KEY6::GetKeyJoy( void )
{
	PRINTD( KEY_LOG, "[KEY][GetKeyJoy]\n" );
	
	int NumKeys;
	Uint8 *KeyST;
	
	KeyST = SDL_GetKeyState( &NumKeys );
	
	return	( KeyST[SDLK_SPACE] ? 0x80:0 ) | ( KeyST[SDLK_LEFT] ? 0x20:0 ) |
			( KeyST[SDLK_RIGHT] ? 0x10:0 ) | ( KeyST[SDLK_DOWN] ? 0x08:0 ) |
			( KeyST[SDLK_UP]    ? 0x04:0 ) | ( KeyST[SDLK_END]  ? 0x02:0 ) |
			( (KeyST[SDLK_LSHIFT] | KeyST[SDLK_RSHIFT]) ? 0x01:0 );
}


////////////////////////////////////////////////////////////////
// ܡɥ󥸥ּ
//
// :	ʤ
// :	BYTE	󥸥
//				KI_KANA:  ʥ⡼ON
//				KI_KKANA: ʥ⡼ON
//				KI_CAPS:  CAPS ON
////////////////////////////////////////////////////////////////
BYTE KEY6::GetKeyIndicator( void )
{
	PRINTD( KEY_LOG, "[KEY][GetKeyIndicator] -> " );
	
	BYTE ret = 0;
	
	// ʥ
	if( ON_KANA ) ret |= ON_KKANA ? KI_KKANA : KI_KANA;
	
	// CAPS
	if( ON_CAPS ) ret |= KI_CAPS;
	
	PRINTD1( KEY_LOG, "%d\n", ret );
	
	return ret;
}


////////////////////////////////////////////////////////////////
// ɤǤSAVE
//
// :	Ini		INI֥ȥݥ
// :	BOOL	TRUE: FALSE:
////////////////////////////////////////////////////////////////
BOOL KEY6::DokoSave( cIni *Ini )
{
	if( !Ini ) return FALSE;
	
	Ini->PutEntry( "KEY", NULL, "ON_KANA",		"%s",	ON_KANA  ? "Yes" : "No" );
	Ini->PutEntry( "KEY", NULL, "ON_KKANA",		"%s",	ON_KKANA ? "Yes" : "No" );
	Ini->PutEntry( "KEY", NULL, "ON_STOP",		"%s",	ON_STOP  ? "Yes" : "No" );
	Ini->PutEntry( "KEY", NULL, "ON_CAPS",		"%s",	ON_CAPS  ? "Yes" : "No" );
	
	if( ak.Buffer && ak.Num > 0 ){
		char strva[strlen(ak.Buffer)*2];
		for( int i=0; i<ak.Num; i++ ) sprintf( &strva[i*2], "%02X", ak.Buffer[i] );
		Ini->PutEntry( "TAPE", NULL, "AKEY_Buffer",		"%s",	strva );
		Ini->PutEntry( "TAPE", NULL, "AKEY_Wait",		"%d",	ak.Wait );
		Ini->PutEntry( "TAPE", NULL, "AKEY_Relay",		"%s",	ak.Relay ? "Yes" : "No" );
		Ini->PutEntry( "TAPE", NULL, "AKEY_Seek",		"%d",	ak.Seek );
	}
	
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// ɤǤLOAD
//
// :	Ini		INI֥ȥݥ
// :	BOOL	TRUE: FALSE:
////////////////////////////////////////////////////////////////
BOOL KEY6::DokoLoad( cIni *Ini )
{
	char strva[MAX_PATH];
	
	if( !Ini ) return FALSE;
	
	Ini->GetTruth( "KEY", "ON_KANA",	&ON_KANA,	ON_KANA );
	Ini->GetTruth( "KEY", "ON_KKANA",	&ON_KKANA,	ON_KKANA );
	Ini->GetTruth( "KEY", "ON_STOP",	&ON_STOP,	ON_STOP );
	Ini->GetTruth( "KEY", "ON_CAPS",	&ON_CAPS,	ON_CAPS );
	
	if( Ini->GetString( "TAPE", "AKEY_Buffer", strva, "" ) ){
		int num = strlen( strva ) / 2;
		char strak[num+1];
		for( int i=0; i<num; i++ ){
			char dt[5] = "0x";
			strncpy( &dt[2], &strva[i*2], 2 );
			strak[i] = strtol( dt, NULL, 16 );
		}
		SetAutoKey( strak, num );
		Ini->GetInt(   "TAPE",	"AKEY_Wait",		&ak.Wait,		ak.Wait );
		Ini->GetTruth( "TAPE",	"AKEY_Relay",		&ak.Relay,		ak.Relay );
		Ini->GetInt(   "TAPE",	"AKEY_Seek",		&ak.Seek,		ak.Seek );
	}
	
	return TRUE;
}
