#include <stdlib.h>
#include <string.h>
#include <new>

#include "memory.h"
#include "common.h"
#include "error.h"
#include "osd.h"


// ֥å
#define MAIN_ROM_0	(MainRom)
#define MAIN_ROM_1	(MainRom+0x2000)
#define MAIN_ROM_2	(MainRom+0x4000)
#define MAIN_ROM_3	(MainRom+0x6000)
#define EXT_ROM_0	(UseExtRom ? ExtRom        : EmptyMEM)
#define EXT_ROM_1	(UseExtRom ? ExtRom+0x2000 : EmptyMEM)
#define KANJI_ROM_0	(KanjiRom)
#define KANJI_ROM_1	(KanjiRom+0x2000)
#define KANJI_ROM_2	(KanjiRom+0x4000)
#define KANJI_ROM_3	(KanjiRom+0x6000)
#define VOICE_ROM_0	(VoiceRom)
#define VOICE_ROM_1	(VoiceRom+0x2000)
#define MAIN_RAM_0	(MainRam)
#define MAIN_RAM_1	(MainRam+0x2000)
#define MAIN_RAM_2	(MainRam+0x4000)
#define MAIN_RAM_3	(MainRam+0x6000)
#define MAIN_RAM_4	(MainRam+0x8000)
#define MAIN_RAM_5	(MainRam+0xa000)
#define MAIN_RAM_6	(MainRam+0xc000)
#define MAIN_RAM_7	(MainRam+0xe000)
#define EXT_RAM_0	(EmptyMEM)
#define EXT_RAM_1	(EmptyMEM)
#define EXT_RAM_2	(EmptyMEM)
#define EXT_RAM_3	(EmptyMEM)
#define EXT_RAM_4	(EmptyMEM)
#define EXT_RAM_5	(EmptyMEM)
#define EXT_RAM_6	(EmptyMEM)
#define EXT_RAM_7	(EmptyMEM)
#define KJ_ROM_0	(kj_rom ? ( kj_LR ? KANJI_ROM_2 : KANJI_ROM_0 ) : VOICE_ROM_0 )
#define KJ_ROM_1	(KJ_ROM_0+0x2000)

// ꥳȥ
#define	INIT_RF0	(0x71)
#define INIT_RF1	(0xdd)
#define INIT_RF2	(0x50)


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
// 
const MEM6::MEMINFO MEM6::IEXTROM   = { NULL,          0x4000,  0,          0xff };
const MEM6::MEMINFO MEM6::IEMPTY    = { NULL,          0x2000,  0,          0xff };
const MEM6::MEMINFO MEM6::IDUMMY    = { NULL,          0x2000,  0,          0xff };

// PC-6001
const MEM6::MEMINFO MEM60::IMAINROM = { "BASICROM.60", 0x4000,  0x54c03109, 0xff };
const MEM6::MEMINFO MEM60::ICGROM1  = { "CGROM60.60",  0x1000,  0xb0142d32, 0xff };
const MEM6::MEMINFO MEM60::IMAINRAM = { NULL,          0x8000,  0,          0    };

// PC-6001mk2
const MEM6::MEMINFO MEM62::IMAINROM = { "BASICROM.62", 0x8000,  0x950ac401, 0xff };
const MEM6::MEMINFO MEM62::ICGROM1  = { "CGROM60.62",  0x2000,  0x81eb5d95, 0xff };
const MEM6::MEMINFO MEM62::ICGROM2  = { "CGROM60m.62", 0x2000,  0x3ce48c33, 0xff };
const MEM6::MEMINFO MEM62::IKANJI   = { "KANJIROM.62", 0x8000,  0x20c8f3eb, 0xff };
const MEM6::MEMINFO MEM62::IVOICE   = { "VOICEROM.62", 0x4000,  0x49b4f917, 0xff };
const MEM6::MEMINFO MEM62::IMAINRAM = { NULL,          0x10000, 0,          0    };

// PC-6601
const MEM6::MEMINFO MEM66::IMAINROM = { "BASICROM.66", 0x8000,  0xc0b01772, 0xff };
const MEM6::MEMINFO MEM66::ICGROM1  = { "CGROM60.66",  0x2000,  0xd2434f29, 0xff };
const MEM6::MEMINFO MEM66::ICGROM2  = { "CGROM66.66",  0x2000,  0x3ce48c33, 0xff };
const MEM6::MEMINFO MEM66::IKANJI   = { "KANJIROM.66", 0x8000,  0x20c8f3eb, 0xff };
const MEM6::MEMINFO MEM66::IVOICE   = { "VOICEROM.66", 0x4000,  0x91d078c1, 0xff };
const MEM6::MEMINFO MEM66::IMAINRAM = { NULL,          0x10000, 0,          0    };

// PC-6001mk2SR
// PC-6601SR


////////////////////////////////////////////////////////////////
// 󥹥ȥ饯
////////////////////////////////////////////////////////////////
MEM6::MEM6( void )
{
	CGBank       = FALSE;	// CG ROM BANK ̵
	UseExtRom    = FALSE;	// ĥROM̵
	EnableChkCRC = TRUE;	// CRCåͭ
}

MEM60::MEM60( BOOL extram )
{
	useextram = extram;	// ĥRAMȤ?
}

MEM62::MEM62( const ID& id ) : Device(id)
{
	Rf0 = INIT_RF0;	// ꥳȥݡȽ
	Rf1 = INIT_RF1;	// ꥳȥݡȽ
	Rf2 = INIT_RF2;	// ꥳȥݡȽ
	
	cgrom  = TRUE;
	kj_rom = TRUE;
	kj_LR  = TRUE;
}


MEM66::MEM66( const ID& id ) : MEM62(id){}


////////////////////////////////////////////////////////////////
// ǥȥ饯
////////////////////////////////////////////////////////////////
MEM6::~MEM6( void )
{
	if( MainRom )	delete[] MainRom;
	if( MainRam )	delete[] MainRam;
	if( CGRom1 )	delete[] CGRom1;
	if( ExtRom )	delete[] ExtRom;
	if( EmptyMEM )	delete[] EmptyMEM;
	if( DummyRAM )	delete[] DummyRAM;
}

MEM60::~MEM60( void ){}

MEM62::~MEM62( void )
{
	if( CGRom2 )	delete[] CGRom2;
	if( KanjiRom )	delete[] KanjiRom;
	if( VoiceRom )	delete[] VoiceRom;
}

MEM66::~MEM66( void ){}


////////////////////////////////////////////////////////////////
// ĥROM ޥ
////////////////////////////////////////////////////////////////
BOOL MEM6::MountExtRom( char *filename )
{
	PRINTD1( MEM_LOG, "[MEM][MountExtRom] -> %s\n", filename );
	
	DWORD LoadSize = 0;
	
	UnmountExtRom();	// ö
	
	try{
		FILE *fp = fopen( filename, "rb" );
		if( !fp ) throw Error::ExtRomMountFailed;
		LoadSize = fread( ExtRom, sizeof(BYTE), IEXTROM.Size, fp );
		fclose( fp );
		
		// å
		if( LoadSize != IEXTROM.Size ) throw Error::RomSizeNG;
	}
	// 㳰ȯ
	catch( Error::Errno i ){
		Error::SetError( i );
		
		UnmountExtRom();
		return FALSE;
	}
	
	UseExtRom = TRUE;
	Reset();
	
	Error::SetError( Error::NoError );
	return TRUE;
}


////////////////////////////////////////////////////////////////
// ĥROM ޥ
////////////////////////////////////////////////////////////////
void MEM6::UnmountExtRom( void )
{
	PRINTD( MEM_LOG, "[MEM][UnmountExtRom]\n" );
	
	memset( ExtRom, IEXTROM.Init, IEXTROM.Size );
	
	UseExtRom = FALSE;
	Reset();
}


////////////////////////////////////////////////////////////////
// ݤROMեɹ
////////////////////////////////////////////////////////////////
BOOL MEM6::AllocMemory( BYTE **buf, const MEMINFO *info, char *path )
{
	PRINTD1( MEM_LOG, "[MEM][AllocMemory] -> %s\n", info->FileName );
	
	BYTE *buffer = NULL;
	DWORD LoadSize = 0;
	
	try{
		// (Ǿ8KB)
		if( info->Size < 0x2000 ) buffer = new BYTE[0x2000];
		else                      buffer = new BYTE[info->Size];
		memset( buffer, info->Init, info->Size );
		
		*buf = buffer;
		
		// ե̾=NULL ʤRAMޤNULL
		if( !info->FileName ) throw Error::NoError;
		
		// ե뤫ɹ
		char fpath[PATH_MAX] = "";
		if( path ){
			strncpy( fpath, path, PATH_MAX );
			ADDDELIMITER( fpath );
		}
		strncat( fpath, info->FileName, PATH_MAX );
		
		FILE *fp = fopen( fpath, "rb" );
		if( !fp ) throw Error::NoRom;
		LoadSize = fread( buffer, sizeof(BYTE), info->Size, fp );
		fclose( fp );
		
		// å
		if( LoadSize != info->Size ) throw Error::RomSizeNG;
		
		// CRCå
		if( EnableChkCRC && (info->Crc != 0) ){	// EnableChkCRC=FALSE ޤ CRC=0λϥåʤ
			if( CalcCrc32( buffer, info->Size ) !=info->Crc ) throw Error::RomCrcNG;
		}
	}
	// new ˼Ԥ
	catch( std::bad_alloc ){
		Error::SetError( Error::MemAllocFailed );
		return FALSE;
	}
	// 㳰ȯ
	catch( Error::Errno i ){
		Error::SetError( i );
		
		switch(i){
		case Error::NoError:	// ե̾=NULL ξ
			return TRUE;
			
		case Error::NoRom:		// եΥץ˼Ԥ
		case Error::RomSizeNG:	// ʤ
		case Error::RomCrcNG:	// CRCʤ
		default:				// 
			delete [] buffer;
			*buf = NULL;
		}
		return FALSE;
	}
	
	Error::SetError( Error::NoError );
	return TRUE;
}


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
BOOL MEM6::InitCommon( BOOL chkcrc )
{
	PRINTD( MEM_LOG, "[MEM][InitCommon]\n" );
	
	EnableChkCRC = chkcrc;
	
	if( !AllocMemory( &ExtRom,   &IEXTROM, NULL ) ) return FALSE;
	if( !AllocMemory( &EmptyMEM, &IEMPTY,  NULL ) ) return FALSE;
	if( !AllocMemory( &DummyRAM, &IDUMMY,  NULL ) ) return FALSE;
	
	PRINTD1( MEM_LOG, "[MEM][Alloc] ExtRom   :%08X\n", (int)ExtRom );
	PRINTD1( MEM_LOG, "[MEM][Alloc] EmptyMEM :%08X\n", (int)EmptyMEM );
	PRINTD1( MEM_LOG, "[MEM][Alloc] DummyRAM :%08X\n", (int)DummyRAM );
	
	return TRUE;
}

BOOL MEM60::Init( char *path, BOOL chkcrc )
{
	PRINTD( MEM_LOG, "[MEM][Init]\n" );
	
	if( !InitCommon( chkcrc ) ) return FALSE;	// ʬ
	
	if( !AllocMemory( &MainRom, &IMAINROM, path ) ) return FALSE;
	if( !AllocMemory( &CGRom1,  &ICGROM1,  path ) ) return FALSE;
	if( !AllocMemory( &MainRam, &IMAINRAM, path ) ) return FALSE;
	
	Reset();
	
	PRINTD1( MEM_LOG, "[MEM][Alloc] MainRam  :%08X\n", (int)MainRam );
	PRINTD1( MEM_LOG, "[MEM][Alloc] MainRom  :%08X\n", (int)MainRom );
	PRINTD1( MEM_LOG, "[MEM][Alloc] CGRom1   :%08X\n", (int)CGRom1 );
	
	return TRUE;
}

BOOL MEM62::Init( char *path, BOOL chkcrc )
{
	PRINTD( MEM_LOG, "[MEM][Init]\n" );
	
	if( !InitCommon( chkcrc ) ) return FALSE;	// ʬ
	
	if( !AllocMemory( &MainRom,  &IMAINROM, path ) ) return FALSE;
	if( !AllocMemory( &CGRom1,   &ICGROM1,  path ) ) return FALSE;
	if( !AllocMemory( &CGRom2,   &ICGROM2,  path ) ) return FALSE;
	if( !AllocMemory( &KanjiRom, &IKANJI,   path ) ) return FALSE;
	if( !AllocMemory( &VoiceRom, &IVOICE,   path ) ) return FALSE;
	if( !AllocMemory( &MainRam,  &IMAINRAM, path ) ) return FALSE;
	
	Reset();
	
	PRINTD1( MEM_LOG, "[MEM][Alloc] MainRam  :%08X\n", (int)MainRam );
	PRINTD1( MEM_LOG, "[MEM][Alloc] MainRom  :%08X\n", (int)MainRom );
	PRINTD1( MEM_LOG, "[MEM][Alloc] CGRom1   :%08X\n", (int)CGRom1 );
	PRINTD1( MEM_LOG, "[MEM][Alloc] CGRom2   :%08X\n", (int)CGRom2 );
	PRINTD1( MEM_LOG, "[MEM][Alloc] KanjiRom :%08X\n", (int)KanjiRom );
	PRINTD1( MEM_LOG, "[MEM][Alloc] VoiceRom :%08X\n", (int)VoiceRom );
	
	return TRUE;
}

BOOL MEM66::Init( char *path, BOOL chkcrc )
{
	PRINTD( MEM_LOG, "[MEM][Init]\n" );
	
	if( !InitCommon( chkcrc ) ) return FALSE;	// ʬ
	
	if( !AllocMemory( &MainRom,  &IMAINROM, path ) ) return FALSE;
	if( !AllocMemory( &CGRom1,   &ICGROM1,  path ) ) return FALSE;
	if( !AllocMemory( &CGRom2,   &ICGROM2,  path ) ) return FALSE;
	if( !AllocMemory( &KanjiRom, &IKANJI,   path ) ) return FALSE;
	if( !AllocMemory( &VoiceRom, &IVOICE,   path ) ) return FALSE;
	if( !AllocMemory( &MainRam,  &IMAINRAM, path ) ) return FALSE;
	
	Reset();
	
	PRINTD1( MEM_LOG, "[MEM][Alloc] MainRam  :%08X\n", (int)MainRam );
	PRINTD1( MEM_LOG, "[MEM][Alloc] MainRom  :%08X\n", (int)MainRom );
	PRINTD1( MEM_LOG, "[MEM][Alloc] CGRom1   :%08X\n", (int)CGRom1 );
	PRINTD1( MEM_LOG, "[MEM][Alloc] CGRom2   :%08X\n", (int)CGRom2 );
	PRINTD1( MEM_LOG, "[MEM][Alloc] KanjiRom :%08X\n", (int)KanjiRom );
	PRINTD1( MEM_LOG, "[MEM][Alloc] VoiceRom :%08X\n", (int)VoiceRom );
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// ꥻå
////////////////////////////////////////////////////////////////
void MEM60::Reset( void )
{
	CGBank = FALSE;	// CG ROM BANK ̵
	
	// CGROMθȾ4KB˳ĥROMƤ򥳥ԡ
	// (Хڤؤ8KBñ̤ǹԤ)
	memcpy( &CGRom1[0x1000], &ExtRom[0x3000], 0x1000 );
	
	SetMB_60();
}

void MEM62::Reset( void )
{
	CGBank = FALSE;	// CG ROM BANK ̵
	Rf0 = INIT_RF0;	// ꥳȥݡȽ
	Rf1 = INIT_RF1;	// ꥳȥݡȽ
	Rf2 = INIT_RF2;	// ꥳȥݡȽ
	
	SetMB_R( Rf0, Rf1 );
	SetMB_W( Rf2 );
}


////////////////////////////////////////////////////////////////
// ꡼
////////////////////////////////////////////////////////////////
BYTE MEM6::Read( WORD addr )
{
	PRINTD1( MEM_LOG, "[MEM][Read] -> %04X", addr );
	
	return rm_blk[addr>>13][addr&0x1fff];
}


////////////////////////////////////////////////////////////////
// 饤
////////////////////////////////////////////////////////////////
void MEM6::Write( WORD addr, BYTE data )
{
	PRINTD5( MEM_LOG, "[MEM][Write] %04X:%02X -> %08X[%04X]'%c'", addr, data, (int)wm_blk[addr>>13], addr&0x1fff, data );
	
	// ȤꤢRAM̵
//	if( ram_sel[addr>>14]&0x01 ){

		wm_blk[addr>>13][addr&0x1fff] = data;
//	}
}


////////////////////////////////////////////////////////////////
// ꡼ɻΥ֥å(62,66Τ)
////////////////////////////////////////////////////////////////
void MEM62::SetMB_R( BYTE mem1, BYTE mem2 )
{
	PRINTD( MEM_LOG, "[MEM][SetMB_R]\n" );
	
	// Port F0H
	switch( mem1 & 0x0f ){	// RF0 (0x0000 - 0x3fff)
		case 0x00:	rm_blk[0]=EmptyMEM;		rm_blk[1]=EmptyMEM;		break;
		case 0x01:	rm_blk[0]=MAIN_ROM_0;	rm_blk[1]=MAIN_ROM_1;	break;
		case 0x02:	rm_blk[0]=KJ_ROM_0;		rm_blk[1]=KJ_ROM_1;		break;
		case 0x03:	rm_blk[0]=EXT_ROM_1;	rm_blk[1]=EXT_ROM_1;	break;
		case 0x04:	rm_blk[0]=EXT_ROM_0;	rm_blk[1]=EXT_ROM_0;	break;
		case 0x05:	rm_blk[0]=KJ_ROM_0;		rm_blk[1]=MAIN_ROM_1;	break;
		case 0x06:	rm_blk[0]=MAIN_ROM_0;	rm_blk[1]=KJ_ROM_1;		break;
		case 0x07:	rm_blk[0]=EXT_ROM_0;	rm_blk[1]=EXT_ROM_1;	break;
		case 0x08:	rm_blk[0]=EXT_ROM_1;	rm_blk[1]=EXT_ROM_0;	break;
		case 0x09:	rm_blk[0]=EXT_ROM_1;	rm_blk[1]=MAIN_ROM_1;	break;
		case 0x0a:	rm_blk[0]=MAIN_ROM_0;	rm_blk[1]=EXT_ROM_1;	break;
		case 0x0b:	rm_blk[0]=EXT_ROM_0;	rm_blk[1]=KJ_ROM_1;		break;
		case 0x0c:	rm_blk[0]=KJ_ROM_0;		rm_blk[1]=EXT_ROM_0;	break;
		case 0x0d:	rm_blk[0]=MAIN_RAM_0;	rm_blk[1]=MAIN_RAM_1;	break;
		case 0x0e:	rm_blk[0]=EXT_RAM_0;	rm_blk[1]=EXT_RAM_1;	break;
		case 0x0f:	rm_blk[0]=EmptyMEM;		rm_blk[1]=EmptyMEM;		break;
	}
	switch( mem1 & 0xf0 ){	// RF0 (0x4000 - 0x7fff)
		case 0x00:	rm_blk[2]=EmptyMEM;		rm_blk[3]=EmptyMEM;		break;
		case 0x10:	rm_blk[2]=MAIN_ROM_2;	rm_blk[3]=MAIN_ROM_3;	break;
		case 0x20:	rm_blk[2]=KJ_ROM_0;		rm_blk[3]=KJ_ROM_1;		break;
		case 0x30:	rm_blk[2]=EXT_ROM_1;	rm_blk[3]=EXT_ROM_1;	break;
		case 0x40:	rm_blk[2]=EXT_ROM_0;	rm_blk[3]=EXT_ROM_0;	break;
		case 0x50:	rm_blk[2]=KJ_ROM_0;		rm_blk[3]=MAIN_ROM_3;	break;
		case 0x60:	rm_blk[2]=MAIN_ROM_2;	rm_blk[3]=KJ_ROM_1;		break;
		case 0x70:	rm_blk[2]=EXT_ROM_0;	rm_blk[3]=EXT_ROM_1;	break;
		case 0x80:	rm_blk[2]=EXT_ROM_1;	rm_blk[3]=EXT_ROM_0;	break;
		case 0x90:	rm_blk[2]=EXT_ROM_1;	rm_blk[3]=MAIN_ROM_3;	break;
		case 0xa0:	rm_blk[2]=MAIN_ROM_2;	rm_blk[3]=EXT_ROM_1;	break;
		case 0xb0:	rm_blk[2]=EXT_ROM_0;	rm_blk[3]=KJ_ROM_1;		break;
		case 0xc0:	rm_blk[2]=KJ_ROM_0;		rm_blk[3]=EXT_ROM_0;	break;
		case 0xd0:	rm_blk[2]=MAIN_RAM_2;	rm_blk[3]=MAIN_RAM_3;	break;
		case 0xe0:	rm_blk[2]=EXT_RAM_2;	rm_blk[3]=EXT_RAM_3;	break;
		case 0xf0:	rm_blk[2]=EmptyMEM;		rm_blk[3]=EmptyMEM;		break;
	}
	if( CGBank ) rm_blk[3] = cgrom ? CGRom1 : CGRom2;
	
	// Port F1H
	switch( mem2 & 0x0f ){	// RF1 (0x8000 - 0xbfff)
		case 0x00:	rm_blk[4]=EmptyMEM;		rm_blk[5]=EmptyMEM;		break;
		case 0x01:	rm_blk[4]=MAIN_ROM_0;	rm_blk[5]=MAIN_ROM_1;	break;
		case 0x02:	rm_blk[4]=KJ_ROM_0;		rm_blk[5]=KJ_ROM_1;		break;
		case 0x03:	rm_blk[4]=EXT_ROM_1;	rm_blk[5]=EXT_ROM_1;	break;
		case 0x04:	rm_blk[4]=EXT_ROM_0;	rm_blk[5]=EXT_ROM_0;	break;
		case 0x05:	rm_blk[4]=KJ_ROM_0;		rm_blk[5]=MAIN_ROM_1;	break;
		case 0x06:	rm_blk[4]=MAIN_ROM_0;	rm_blk[5]=KJ_ROM_1;		break;
		case 0x07:	rm_blk[4]=EXT_ROM_0;	rm_blk[5]=EXT_ROM_1;	break;
		case 0x08:	rm_blk[4]=EXT_ROM_1;	rm_blk[5]=EXT_ROM_0;	break;
		case 0x09:	rm_blk[4]=EXT_ROM_1;	rm_blk[5]=MAIN_ROM_1;	break;
		case 0x0a:	rm_blk[4]=MAIN_ROM_0;	rm_blk[5]=EXT_ROM_1;	break;
		case 0x0b:	rm_blk[4]=EXT_ROM_0;	rm_blk[5]=KJ_ROM_1;		break;
		case 0x0c:	rm_blk[4]=KJ_ROM_0;		rm_blk[5]=EXT_ROM_0;	break;
		case 0x0d:	rm_blk[4]=MAIN_RAM_4;	rm_blk[5]=MAIN_RAM_5;	break;
		case 0x0e:	rm_blk[4]=EXT_RAM_4;	rm_blk[5]=EXT_RAM_5;	break;
		case 0x0f:	rm_blk[4]=EmptyMEM;		rm_blk[5]=EmptyMEM;		break;
	}
	switch( mem2 & 0xf0 ){	// RF1 (0xc000 - 0xffff)
		case 0x00:	rm_blk[6]=EmptyMEM;		rm_blk[7]=EmptyMEM;		break;
		case 0x10:	rm_blk[6]=MAIN_ROM_2;	rm_blk[7]=MAIN_ROM_3;	break;
		case 0x20:	rm_blk[6]=KJ_ROM_0;		rm_blk[7]=KJ_ROM_1;		break;
		case 0x30:	rm_blk[6]=EXT_ROM_1;	rm_blk[7]=EXT_ROM_1;	break;
		case 0x40:	rm_blk[6]=EXT_ROM_0;	rm_blk[7]=EXT_ROM_0;	break;
		case 0x50:	rm_blk[6]=KJ_ROM_0;		rm_blk[7]=MAIN_ROM_3;	break;
		case 0x60:	rm_blk[6]=MAIN_ROM_2;	rm_blk[7]=KJ_ROM_1;		break;
		case 0x70:	rm_blk[6]=EXT_ROM_0;	rm_blk[7]=EXT_ROM_1;	break;
		case 0x80:	rm_blk[6]=EXT_ROM_1;	rm_blk[7]=EXT_ROM_0;	break;
		case 0x90:	rm_blk[6]=EXT_ROM_1;	rm_blk[7]=MAIN_ROM_3;	break;
		case 0xa0:	rm_blk[6]=MAIN_ROM_2;	rm_blk[7]=EXT_ROM_1;	break;
		case 0xb0:	rm_blk[6]=EXT_ROM_0;	rm_blk[7]=KJ_ROM_1;		break;
		case 0xc0:	rm_blk[6]=KJ_ROM_0;		rm_blk[7]=EXT_ROM_0;	break;
		case 0xd0:	rm_blk[6]=MAIN_RAM_6;	rm_blk[7]=MAIN_RAM_7;	break;
		case 0xe0:	rm_blk[6]=EXT_RAM_6;	rm_blk[7]=EXT_RAM_7;	break;
		case 0xf0:	rm_blk[6]=EmptyMEM;		rm_blk[7]=EmptyMEM;		break;
	}
	
	// ¸
	Rf0 = mem1;
	Rf1 = mem2;
	
	#if (MEM_LOG)
		char mb[32];
		fprintf( stdout, "[MEM][SetMB_R] %02X %02X\n", mem1, mem2 );
		for( int i=0; i<8; i++ ) fprintf( stdout, "               %d:%8s\n", i, GetReadMemBlk(i,mb) );
	#endif
}



////////////////////////////////////////////////////////////////
// 饤ȻΥ֥å(62,66Τ)
////////////////////////////////////////////////////////////////
void MEM62::SetMB_W( BYTE mem )
{
	PRINTD1( MEM_LOG, "[MEM][SetMB_W] -> %02X\n", mem );
	
	if( (ram_sel[0]=(mem   )&1) ){ wm_blk[0]=MAIN_RAM_0;	wm_blk[1]=MAIN_RAM_1; }
	else                         { wm_blk[0]=DummyRAM;		wm_blk[1]=DummyRAM;   }
	if( (ram_sel[1]=(mem>>2)&1) ){ wm_blk[2]=MAIN_RAM_2;	wm_blk[3]=MAIN_RAM_3; }
	else                         { wm_blk[2]=DummyRAM;		wm_blk[3]=DummyRAM;   }
	if( (ram_sel[2]=(mem>>4)&1) ){ wm_blk[4]=MAIN_RAM_4;	wm_blk[5]=MAIN_RAM_5; }
	else                         { wm_blk[4]=DummyRAM;		wm_blk[5]=DummyRAM;   }
	if( (ram_sel[3]=(mem>>6)&1) ){ wm_blk[6]=MAIN_RAM_6;	wm_blk[7]=MAIN_RAM_7; }
	else                         { wm_blk[6]=DummyRAM;		wm_blk[7]=DummyRAM;   }
	
	// ¸
	Rf2 = mem;
	
	#if (MEM_LOG)
		char mb[32];
		for( int i=0; i<8; i++ ) fprintf( stdout, "               %d:%8s\n", i, GetWriteMemBlk(i,mb) );
	#endif
}


////////////////////////////////////////////////////////////////
// ֥å(60Τ)
////////////////////////////////////////////////////////////////
void MEM60::SetMB_60( void )
{
	PRINTD( MEM_LOG, "[MEM][SetMB_60]\n" );
	
	// Read
	rm_blk[0]=MAIN_ROM_0; rm_blk[1]=MAIN_ROM_1;
	rm_blk[2]=EXT_ROM_0;  rm_blk[3]=EXT_ROM_1;
	// ĥRAMȤ?
	if( useextram ){ rm_blk[4]=MAIN_RAM_0; rm_blk[5]=MAIN_RAM_1; }
	else           { rm_blk[4]=EmptyMEM;   rm_blk[5]=EmptyMEM;   }
	rm_blk[6]=MAIN_RAM_2; rm_blk[7]=MAIN_RAM_3;
	
	// Write
	ram_sel[0]=0; wm_blk[0]=DummyRAM; wm_blk[1]=DummyRAM;
	ram_sel[1]=0; wm_blk[2]=DummyRAM; wm_blk[3]=DummyRAM;
	// ĥRAMȤ?
	if( useextram ){ ram_sel[2]=1; wm_blk[4]=MAIN_RAM_0; wm_blk[5]=MAIN_RAM_1; }
	else           { ram_sel[2]=0; wm_blk[4]=DummyRAM;   wm_blk[5]=DummyRAM;   }
	ram_sel[3]=1; wm_blk[6]=MAIN_RAM_2; wm_blk[7]=MAIN_RAM_3;
}


////////////////////////////////////////////////////////////////
// CG ROM (62,66Τ)
////////////////////////////////////////////////////////////////
void MEM62::SelectCGrom( int mode )
{
	PRINTD1( MEM_LOG, "[MEM][SelectCGrom] -> %d\n", mode );
	
	// mode 1:32*16(N60⡼) 0:40*20(N60m⡼)
	cgrom = mode ? TRUE : FALSE;
}


////////////////////////////////////////////////////////////////
// CG ROM BANK ڤؤ
////////////////////////////////////////////////////////////////
void MEM6::SetCGBank( BOOL data )
{
	PRINTD1( MEM_LOG, "[MEM][SetCGBank] -> %d\n", data );
	
	CGBank = data;
	rm_blk[3] = CGBank ? CGRom1 : EXT_ROM_1;
}

void MEM62::SetCGBank( BOOL data )
{
	PRINTD1( MEM_LOG, "[MEM][SetCGBank] -> %d\n", data );
	
	CGBank = data;
	SetMB_R( Rf0, Rf1 );
}


////////////////////////////////////////////////////////////////
// ROM  ROM ڤؤ(62,66Τ)
////////////////////////////////////////////////////////////////
void MEM62::SelectKanjiRom( BYTE mode )
{
	PRINTD1( MEM_LOG, "[MEM][SelectKanjiRom] -> %02X\n", mode );
	
// mode bit0 0:ROM 1:ROM
//      bit1 0:ROM¦     1:ROM¦

	//  
	kj_LR = mode&0x02 ? TRUE : FALSE;
	
	// ROMROM
	kj_rom = mode&0x01 ? TRUE : FALSE;
	
	SetMB_R( Rf0, Rf1 );
}

void MEM66::SelectKanjiRom( BYTE mode )
{
	MEM62::SelectKanjiRom( mode );
	
	// 66ξ,äȰ㤦餷
	if( !(mode&0x01) ){
		if( rm_blk[0] != MAIN_ROM_0 ) rm_blk[0]=VOICE_ROM_0;
		if( rm_blk[1] != MAIN_ROM_1 ) rm_blk[1]=VOICE_ROM_1;
	}
}


////////////////////////////////////////////////////////////////
// ɤǤSAVE
////////////////////////////////////////////////////////////////
BOOL MEM60::DokoSave( cIni *Ini )
{
	if( !Ini ) return FALSE;
	
	for( int i=0; i<0x8000; i+=64 ){
		char stren[16],strva[256];
		sprintf( stren, "MainRam_%04X", i );
		for( int j=0; j<64; j++ ) sprintf( &strva[j*2], "%02X", MainRam[i+j] );
		Ini->PutEntry( "MEMORY", NULL, stren, "%s", strva );
	}
	
	Ini->PutEntry( "MEMORY", NULL, "CGBank",	"%s",	CGBank ? "Yes" : "No" );
	Ini->PutEntry( "MEMORY", NULL, "useextram",	"%s",	useextram ? "Yes" : "No" );
	
	return TRUE;
}

BOOL MEM62::DokoSave( cIni *Ini )
{
	if( !Ini ) return FALSE;
	
	for( int i=0; i<0x10000; i+=64 ){
		char stren[16],strva[256];
		sprintf( stren, "MainRam_%04X", i );
		for( int j=0; j<64; j++ ) sprintf( &strva[j*2], "%02X", MainRam[i+j] );
		Ini->PutEntry( "MEMORY", NULL, stren, "%s", strva );
	}
	
	Ini->PutEntry( "MEMORY", NULL, "CGBank",	"%s",	CGBank ? "Yes" : "No" );
	Ini->PutEntry( "MEMORY", NULL, "cgrom",		"%s",	cgrom ? "Yes" : "No" );
	Ini->PutEntry( "MEMORY", NULL, "kj_rom",	"%s",	kj_rom ? "Yes" : "No" );
	Ini->PutEntry( "MEMORY", NULL, "kj_LR",		"%s",	kj_LR ? "Yes" : "No" );
	Ini->PutEntry( "MEMORY", NULL, "Rf0",		"%d",	Rf0 );
	Ini->PutEntry( "MEMORY", NULL, "Rf1",		"%d",	Rf1 );
	Ini->PutEntry( "MEMORY", NULL, "Rf2",		"%d",	Rf2 );
	
	return TRUE;
}



////////////////////////////////////////////////////////////////
// ɤǤLOAD
////////////////////////////////////////////////////////////////
BOOL MEM60::DokoLoad( cIni *Ini )
{
	if( !Ini ) return FALSE;
	
	for( int i=0; i<0x8000; i+=64 ){
		char stren[16],strva[256];
		sprintf( stren, "MainRam_%04X", i );
		memset( strva, '0', 64*2 );
		Ini->GetString( "MEMORY", stren, strva, strva );
		
		for( int j=0; j<64; j++ ){
			char dt[5] = "0x";
			strncpy( &dt[2], &strva[j*2], 2 );
			MainRam[i+j] = strtol( dt, NULL, 16 );
		}
	}
	
	Ini->GetTruth( "MEMORY", "CGBank",		&CGBank,	CGBank );
	Ini->GetTruth( "MEMORY", "useextram",	&useextram,	useextram );
	
	SetMB_60();
	SetCGBank( CGBank );
	
	return TRUE;
}

BOOL MEM62::DokoLoad( cIni *Ini )
{
	int st;
	
	if( !Ini ) return FALSE;
	
	for( int i=0; i<0x10000; i+=64 ){
		char stren[16],strva[256];
		sprintf( stren, "MainRam_%04X", i );
		memset( strva, '0', 64*2 );
		Ini->GetString( "MEMORY", stren, strva, strva );
		
		for( int j=0; j<64; j++ ){
			char dt[5] = "0x";
			strncpy( dt, &strva[j*2], 2 );
			MainRam[i+j] = strtol( dt, NULL, 16 );
		}
	}
	
	Ini->GetTruth( "MEMORY", "CGBank",	&CGBank,	CGBank );
	Ini->GetTruth( "MEMORY", "cgrom",	&cgrom,		cgrom );
	Ini->GetTruth( "MEMORY", "kj_rom",	&kj_rom,	kj_rom );
	Ini->GetTruth( "MEMORY", "kj_LR",	&kj_LR,		kj_LR );
	Ini->GetInt(   "MEMORY", "Rf0",		&st,		Rf0 );	Rf0 = st;
	Ini->GetInt(   "MEMORY", "Rf1",		&st,		Rf1 );	Rf1 = st;
	Ini->GetInt(   "MEMORY", "Rf2",		&st,		Rf2 );	Rf2 = st;
	
	SetMB_R( Rf0, Rf1 );
	SetMB_W( Rf2 );
	SetCGBank( CGBank );
	
	return TRUE;
}


#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
////////////////////////////////////////////////////////////////
// ǥХåѴؿ
////////////////////////////////////////////////////////////////
char *MEM60::GetReadMemBlk( int blk, char *name  )
{
	BYTE *addr = rm_blk[blk];
	
	if     ( addr == EmptyMEM )  { strcpy( name, "EmptyMEM" ); }
	else if( addr == DummyRAM )  { strcpy( name, "DummyRAM" ); }
	else if( addr == MAIN_ROM_0 ){ strcpy( name, "MainRom0" ); }
	else if( addr == MAIN_ROM_1 ){ strcpy( name, "MainRom1" ); }
	else if( addr == MAIN_ROM_2 ){ strcpy( name, "MainRom2" ); }
	else if( addr == MAIN_ROM_3 ){ strcpy( name, "MainRom3" ); }
	else if( addr == EXT_ROM_0 ) { strcpy( name, "ExtRom0" );  }
	else if( addr == EXT_ROM_1 ) { strcpy( name, "ExtRom1" );  }
	else if( addr == MAIN_RAM_0 ){ strcpy( name, "MainRam0" ); }
	else if( addr == MAIN_RAM_1 ){ strcpy( name, "MainRam1" ); }
	else if( addr == MAIN_RAM_2 ){ strcpy( name, "MainRam2" ); }
	else if( addr == MAIN_RAM_3 ){ strcpy( name, "MainRam3" ); }
	else if( addr == MAIN_RAM_4 ){ strcpy( name, "MainRam4" ); }
	else if( addr == MAIN_RAM_5 ){ strcpy( name, "MainRam5" ); }
	else if( addr == MAIN_RAM_6 ){ strcpy( name, "MainRam6" ); }
	else if( addr == MAIN_RAM_7 ){ strcpy( name, "MainRam7" ); }
	else if( addr == EXT_RAM_0 ) { strcpy( name, "ExtRam0" );  }
	else if( addr == EXT_RAM_1 ) { strcpy( name, "ExtRam1" );  }
	else if( addr == EXT_RAM_2 ) { strcpy( name, "ExtRam2" );  }
	else if( addr == EXT_RAM_3 ) { strcpy( name, "ExtRam3" );  }
	else if( addr == EXT_RAM_4 ) { strcpy( name, "ExtRam4" );  }
	else if( addr == EXT_RAM_5 ) { strcpy( name, "ExtRam5" );  }
	else if( addr == EXT_RAM_6 ) { strcpy( name, "ExtRam6" );  }
	else if( addr == EXT_RAM_7 ) { strcpy( name, "ExtRam7" );  }
	
	return name;
}

char *MEM60::GetWriteMemBlk( int blk, char *name  )
{
	BYTE *addr = wm_blk[blk];
	
	if     ( addr == EmptyMEM )  { strcpy( name, "EmptyMEM" ); }
	else if( addr == DummyRAM )  { strcpy( name, "DummyRAM" ); }
	else if( addr == MAIN_ROM_0 ){ strcpy( name, "MainRom0" ); }
	else if( addr == MAIN_ROM_1 ){ strcpy( name, "MainRom1" ); }
	else if( addr == MAIN_ROM_2 ){ strcpy( name, "MainRom2" ); }
	else if( addr == MAIN_ROM_3 ){ strcpy( name, "MainRom3" ); }
	else if( addr == EXT_ROM_0 ) { strcpy( name, "ExtRom0" );  }
	else if( addr == EXT_ROM_1 ) { strcpy( name, "ExtRom1" );  }
	else if( addr == MAIN_RAM_0 ){ strcpy( name, "MainRam0" ); }
	else if( addr == MAIN_RAM_1 ){ strcpy( name, "MainRam1" ); }
	else if( addr == MAIN_RAM_2 ){ strcpy( name, "MainRam2" ); }
	else if( addr == MAIN_RAM_3 ){ strcpy( name, "MainRam3" ); }
	else if( addr == MAIN_RAM_4 ){ strcpy( name, "MainRam4" ); }
	else if( addr == MAIN_RAM_5 ){ strcpy( name, "MainRam5" ); }
	else if( addr == MAIN_RAM_6 ){ strcpy( name, "MainRam6" ); }
	else if( addr == MAIN_RAM_7 ){ strcpy( name, "MainRam7" ); }
	else if( addr == EXT_RAM_0 ) { strcpy( name, "ExtRam0" );  }
	else if( addr == EXT_RAM_1 ) { strcpy( name, "ExtRam1" );  }
	else if( addr == EXT_RAM_2 ) { strcpy( name, "ExtRam2" );  }
	else if( addr == EXT_RAM_3 ) { strcpy( name, "ExtRam3" );  }
	else if( addr == EXT_RAM_4 ) { strcpy( name, "ExtRam4" );  }
	else if( addr == EXT_RAM_5 ) { strcpy( name, "ExtRam5" );  }
	else if( addr == EXT_RAM_6 ) { strcpy( name, "ExtRam6" );  }
	else if( addr == EXT_RAM_7 ) { strcpy( name, "ExtRam7" );  }
	
	return name;
}


char *MEM62::GetReadMemBlk( int blk, char *name  )
{
	BYTE *addr = rm_blk[blk];
	
	if     ( addr == EmptyMEM )   { strcpy( name, "EmptyMEM" ); }
	else if( addr == DummyRAM )   { strcpy( name, "DummyRAM" ); }
	else if( addr == MAIN_ROM_0 ) { strcpy( name, "MainRom0" ); }
	else if( addr == MAIN_ROM_1 ) { strcpy( name, "MainRom1" ); }
	else if( addr == MAIN_ROM_2 ) { strcpy( name, "MainRom2" ); }
	else if( addr == MAIN_ROM_3 ) { strcpy( name, "MainRom3" ); }
	else if( addr == EXT_ROM_0 )  { strcpy( name, "ExtRom0" );  }
	else if( addr == EXT_ROM_1 )  { strcpy( name, "ExtRom1" );  }
	else if( addr == KANJI_ROM_0 ){ strcpy( name, "KanjRom0" ); }
	else if( addr == KANJI_ROM_1 ){ strcpy( name, "KanjRom1" ); }
	else if( addr == KANJI_ROM_2 ){ strcpy( name, "KanjRom2" ); }
	else if( addr == KANJI_ROM_3 ){ strcpy( name, "KanjRom3" ); }
	else if( addr == VOICE_ROM_0 ){ strcpy( name, "VoicRom0" ); }
	else if( addr == VOICE_ROM_1 ){ strcpy( name, "VoicRom1" ); }
	else if( addr == MAIN_RAM_0 ) { strcpy( name, "MainRam0" ); }
	else if( addr == MAIN_RAM_1 ) { strcpy( name, "MainRam1" ); }
	else if( addr == MAIN_RAM_2 ) { strcpy( name, "MainRam2" ); }
	else if( addr == MAIN_RAM_3 ) { strcpy( name, "MainRam3" ); }
	else if( addr == MAIN_RAM_4 ) { strcpy( name, "MainRam4" ); }
	else if( addr == MAIN_RAM_5 ) { strcpy( name, "MainRam5" ); }
	else if( addr == MAIN_RAM_6 ) { strcpy( name, "MainRam6" ); }
	else if( addr == MAIN_RAM_7 ) { strcpy( name, "MainRam7" ); }
	else if( addr == EXT_RAM_0 )  { strcpy( name, "ExtRam0" );  }
	else if( addr == EXT_RAM_1 )  { strcpy( name, "ExtRam1" );  }
	else if( addr == EXT_RAM_2 )  { strcpy( name, "ExtRam2" );  }
	else if( addr == EXT_RAM_3 )  { strcpy( name, "ExtRam3" );  }
	else if( addr == EXT_RAM_4 )  { strcpy( name, "ExtRam4" );  }
	else if( addr == EXT_RAM_5 )  { strcpy( name, "ExtRam5" );  }
	else if( addr == EXT_RAM_6 )  { strcpy( name, "ExtRam6" );  }
	else if( addr == EXT_RAM_7 )  { strcpy( name, "ExtRam7" );  }
	
	return name;
}

char *MEM62::GetWriteMemBlk( int blk, char *name  )
{
	BYTE *addr = wm_blk[blk];
	
	if     ( addr == EmptyMEM )   { strcpy( name, "EmptyMEM" ); }
	else if( addr == DummyRAM )   { strcpy( name, "DummyRAM" ); }
	else if( addr == MAIN_ROM_0 ) { strcpy( name, "MainRom0" ); }
	else if( addr == MAIN_ROM_1 ) { strcpy( name, "MainRom1" ); }
	else if( addr == MAIN_ROM_2 ) { strcpy( name, "MainRom2" ); }
	else if( addr == MAIN_ROM_3 ) { strcpy( name, "MainRom3" ); }
	else if( addr == EXT_ROM_0 )  { strcpy( name, "ExtRom0" );  }
	else if( addr == EXT_ROM_1 )  { strcpy( name, "ExtRom1" );  }
	else if( addr == KANJI_ROM_0 ){ strcpy( name, "KanjRom0" ); }
	else if( addr == KANJI_ROM_1 ){ strcpy( name, "KanjRom1" ); }
	else if( addr == KANJI_ROM_2 ){ strcpy( name, "KanjRom2" ); }
	else if( addr == KANJI_ROM_3 ){ strcpy( name, "KanjRom3" ); }
	else if( addr == VOICE_ROM_0 ){ strcpy( name, "VoicRom0" ); }
	else if( addr == VOICE_ROM_1 ){ strcpy( name, "VoicRom1" ); }
	else if( addr == MAIN_RAM_0 ) { strcpy( name, "MainRam0" ); }
	else if( addr == MAIN_RAM_1 ) { strcpy( name, "MainRam1" ); }
	else if( addr == MAIN_RAM_2 ) { strcpy( name, "MainRam2" ); }
	else if( addr == MAIN_RAM_3 ) { strcpy( name, "MainRam3" ); }
	else if( addr == MAIN_RAM_4 ) { strcpy( name, "MainRam4" ); }
	else if( addr == MAIN_RAM_5 ) { strcpy( name, "MainRam5" ); }
	else if( addr == MAIN_RAM_6 ) { strcpy( name, "MainRam6" ); }
	else if( addr == MAIN_RAM_7 ) { strcpy( name, "MainRam7" ); }
	else if( addr == EXT_RAM_0 )  { strcpy( name, "ExtRam0" );  }
	else if( addr == EXT_RAM_1 )  { strcpy( name, "ExtRam1" );  }
	else if( addr == EXT_RAM_2 )  { strcpy( name, "ExtRam2" );  }
	else if( addr == EXT_RAM_3 )  { strcpy( name, "ExtRam3" );  }
	else if( addr == EXT_RAM_4 )  { strcpy( name, "ExtRam4" );  }
	else if( addr == EXT_RAM_5 )  { strcpy( name, "ExtRam5" );  }
	else if( addr == EXT_RAM_6 )  { strcpy( name, "ExtRam6" );  }
	else if( addr == EXT_RAM_7 )  { strcpy( name, "ExtRam7" );  }
	
	return name;
}
#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //



////////////////////////////////////////////////////////////////
//  device description
////////////////////////////////////////////////////////////////
const Device::Descriptor MEM62::descriptor = {
	MEM62::indef, MEM62::outdef
};

const Device::OutFuncPtr MEM62::outdef[] = {
	STATIC_CAST( Device::OutFuncPtr, &MEM62::OutC1H ),
	STATIC_CAST( Device::OutFuncPtr, &MEM62::OutC2H ),
	STATIC_CAST( Device::OutFuncPtr, &MEM62::OutC3H ),
	STATIC_CAST( Device::OutFuncPtr, &MEM62::OutF0H ),
	STATIC_CAST( Device::OutFuncPtr, &MEM62::OutF1H ),
	STATIC_CAST( Device::OutFuncPtr, &MEM62::OutF2H )
};

const Device::InFuncPtr MEM62::indef[] = {
	STATIC_CAST( Device::InFuncPtr, &MEM62::InF0H ),
	STATIC_CAST( Device::InFuncPtr, &MEM62::InF1H ),
	STATIC_CAST( Device::InFuncPtr, &MEM62::InF2H )
};
