#ifndef MEMORY_H_INCLUDED
#define MEMORY_H_INCLUDED

#include "pc6001v.h"
#include "device.h"
#include "ini.h"


////////////////////////////////////////////////////////////////
// 饹
////////////////////////////////////////////////////////////////
class MEM6 : public IDoko {
protected:
	// ¤
	typedef struct{
		char *FileName;	// ե̾
		Uint32 Size;	// 
		Uint32 Crc;		// CRC32
		BYTE Init;		// ǡ
	} MEMINFO;
	
	// 
	static const MEMINFO IEXTROM;
	static const MEMINFO IEMPTY;
	static const MEMINFO IDUMMY;
	
	BOOL CGBank;		// CG ROM BANK TRUE:ͭ FALSE:̵
	BOOL UseExtRom;		// ĥROM     TRUE:ͭ FALSE:̵
	
	BYTE *MainRom;		// BASIC ROM	60:16KB 62,66:32KB
	BYTE *ExtRom;		// ĥ ROM		60:16KB 62,66:16KB
	BYTE *CGRom1;		// CG ROM1		60: 4KB 62,66: 8KB
	BYTE *MainRam;		// RAM			60:32KB 62,66:64KB
	BYTE *EmptyMEM;		// ̤ 60: 8KB 62,66: 8KB
	BYTE *DummyRAM;		// ߡRAM	60: 8KB 62,66: 8KB
	
	BYTE *rm_blk[8];	// ꡼ɻ֥åݥ(8KB*8)
	BYTE *wm_blk[8];	// 饤Ȼ֥åݥ(8KB*8)
	BYTE ram_sel[4];	// RAM,RAM (16KB*4)
	
	BOOL AllocMemory( BYTE **, const MEMINFO *, char * );	// ݤROMեɹ
	
	BOOL InitCommon( BOOL );					// ʬ
	
	BOOL EnableChkCRC;							// CRCåͭ?
	
public:
	MEM6();										// 󥹥ȥ饯
	virtual ~MEM6();							// ǥȥ饯
	
	virtual BOOL Init( char *, BOOL ) = 0;		// 
	virtual void Reset(){};						// ꥻå
	
	BYTE Read( WORD );							// ꡼
	void Write( WORD, BYTE );					// 饤
	
	BOOL MountExtRom( char * );					// ĥROM ޥ
	void UnmountExtRom();						// ĥROM ޥ
	
	// 8255ϴϢؿ
	virtual void SetCGBank( BOOL );				// CG ROM BANK ڤؤ
	
	// ľܥؿ
	virtual BYTE ReadMainRom( WORD addr ){ return MainRom[addr&0x3fff]; }
	virtual BYTE ReadMainRam( WORD addr ){ return MainRam[addr&0x3fff]; }
	virtual BYTE ReadExtRom ( WORD addr ){ return ExtRom ? ExtRom[addr&0x3fff] : 0xff; }
	virtual BYTE ReadCGrom1 ( WORD addr ){ return CGRom1[addr&0x1fff]; }
	virtual BYTE ReadCGrom2 ( WORD )     { return 0xff; }
	
	#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	virtual char *GetReadMemBlk( int, char * )  = 0;
	virtual char *GetWriteMemBlk( int, char * ) = 0;
	#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
};


class MEM60 : public MEM6 {
private:
	// 
	static const MEMINFO IMAINROM;
	static const MEMINFO ICGROM1;
	static const MEMINFO IMAINRAM;
	
	BOOL useextram;						// ĥRAMȤ?
	
	void SetMB_60();					// ֥å(60Τ)
	
public:
	MEM60( BOOL );						// 󥹥ȥ饯
	~MEM60();							// ǥȥ饯
	
	BOOL Init( char *, BOOL );			// 
	void Reset();						// ꥻå
	
	// ľܥؿ
	BYTE ReadMainRam( WORD addr ){ return MainRam[addr&(useextram ? 0x7fff : 0x3fff)]; }
	
	#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	char *GetReadMemBlk( int, char * );
	char *GetWriteMemBlk( int, char * );
	#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	
	// ------------------------------------------
	BOOL DokoSave( cIni * );	// ɤǤSAVE
	BOOL DokoLoad( cIni * );	// ɤǤLOAD
	// ------------------------------------------
};


class MEM62 : public MEM6, public Device {
protected:
	// 
	static const MEMINFO IMAINROM;
	static const MEMINFO ICGROM1;
	static const MEMINFO ICGROM2;
	static const MEMINFO IKANJI;
	static const MEMINFO IVOICE;
	static const MEMINFO IMAINRAM;
	
	BYTE *CGRom2;		// CG ROM2		60: -KB 62,66: 8KB
	BYTE *KanjiRom;		//  ROM		60: -KB 62,66:32KB
	BYTE *VoiceRom;		//  ROM	60: -KB 62,66:16KB
	
	BOOL cgrom;			// CG ROM  TRUE:N60⡼ FALSE:N60m⡼
	BOOL kj_rom;		// ROM,ROM  TRUE:ROM FALSE:ROM
	BOOL kj_LR;			// ROM  TRUE: FALSE:
	
	BYTE Rf0;			// ꥳȥݡ
	BYTE Rf1;			// ꥳȥݡ
	BYTE Rf2;			// ꥳȥݡ
	
	void SetMB_R( BYTE, BYTE );			// ꡼ɻΥ֥å(62,66Τ)
	void SetMB_W( BYTE );				// 饤ȻΥ֥å(62,66Τ)
	
	void SelectCGrom( int );			// CG ROM (62,66Τ)
	void SelectKanjiRom( BYTE );		// ROM  ROM ڤؤ(62,66Τ)
	
	// ǥХ
	static const Descriptor descriptor;
	static const InFuncPtr  indef[];
	static const OutFuncPtr outdef[];
	const Descriptor* GetDesc() const { return &descriptor; }
	
	// I/Oؿ
	inline void OutC1H( int, BYTE data ){ SelectCGrom( (data>>1)&1 ); }
	inline void OutC2H( int, BYTE data ){ SelectKanjiRom( data ); }
	inline void OutC3H( int, BYTE data ){}
	inline void OutF0H( int, BYTE data ){ SetMB_R( data, Rf1 ); }
	inline void OutF1H( int, BYTE data ){ SetMB_R( Rf0, data ); }
	inline void OutF2H( int, BYTE data ){ SetMB_W( data ); }
	inline BYTE InF0H( int ){ return Rf0; }
	inline BYTE InF1H( int ){ return Rf1; }
	inline BYTE InF2H( int ){ return Rf2; }
	
public:
	MEM62( const ID& );					// 󥹥ȥ饯
	virtual ~MEM62();					// ǥȥ饯
	
	virtual BOOL Init( char *, BOOL );	// 
	void Reset();						// ꥻå
	
	// 8255ϴϢؿ
	void SetCGBank( BOOL );				// CG ROM BANK ڤؤ
	
	// ľܥؿ
	BYTE ReadMainRom( WORD addr ) { return MainRom[addr&0x7fff]; }
	BYTE ReadMainRam( WORD addr ) { return MainRam[addr&0xffff]; }
	BYTE ReadCGrom2( WORD addr )  { return CGRom2[addr&0x1fff]; }
	BYTE ReadKanjiRom( WORD addr ){ return KanjiRom[addr&0x7fff]; }
	BYTE ReadVoiceRom( WORD addr ){ return VoiceRom[addr&0x3fff]; }
	
	#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	char *GetReadMemBlk( int, char * );
	char *GetWriteMemBlk( int, char * );
	#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	
	// ǥХID
	enum IDOut{ outC1H=0, outC2H, outC3H, outF0H, outF1H, outF2H };
	enum IDIn {  inF0H=0,  inF1H,  inF2H };
	
	// ------------------------------------------
	BOOL DokoSave( cIni * );	// ɤǤSAVE
	BOOL DokoLoad( cIni * );	// ɤǤLOAD
	// ------------------------------------------
};


class MEM66 : public MEM62 {
private:
	// 
	static const MEMINFO IMAINROM;
	static const MEMINFO ICGROM1;
	static const MEMINFO ICGROM2;
	static const MEMINFO IKANJI;
	static const MEMINFO IVOICE;
	static const MEMINFO IMAINRAM;
	
	void SelectKanjiRom( BYTE );		// ROM  ROM ڤؤ(62,66Τ)
	
public:
	MEM66( const ID& );					// 󥹥ȥ饯
	~MEM66();							// ǥȥ饯
	
	BOOL Init( char *, BOOL );			// 
};




/*
class MemBlock {
protected:
	BYTE MomoryBlock[0x2000];	// 8KBΥ֥å
	int Wait;					// 
	
public:
	MemBlock( BYTE );			// 󥹥ȥ饯
	virtual ~MemBlock();		// ǥȥ饯
	
	BYTE Read( WORD addr ){ return MomoryBlock[addr&0x1fff]; }	// ꡼
	virtual void Write( WORD, BYTE ){};							// 饤
	
	void SetWait( int wait ){ Wait = wait; }					// 
};

class RomBlock : public MemBlock {
protected:
	
public:
	RomBlock();		// 󥹥ȥ饯
	~RomBlock();	// ǥȥ饯
	
	BOOL LoadRom( char *, BYTE *, DWORD );		// ROMեɹߥ
};

class RamBlock : public MemBlock {
protected:
	
public:
	RamBlock();		// 󥹥ȥ饯
	~RamBlock();	// ǥȥ饯
	
	void Write( WORD addr , BYTE data ){ MomoryBlock[addr&0x1fff] = data; }	// 饤
};
*/


#endif	// MEMORY_H_INCLUDED
