#include <new>

#include "pc60.h"
#include "error.h"

#include "ini.h"

#include "ay8910.h"
#include "breakpoint.h"
#include "debug.h"
#include "disk.h"
#include "graph.h"
#include "intr.h"
#include "io.h"
#include "keyboard.h"
#include "memory.h"
#include "osd.h"
#include "pd8049.h"
#include "pd8255.h"
#include "schedule.h"
#include "sound.h"
#include "status.h"
#include "tape.h"
#include "voice.h"
#include "z80.h"


////////////////////////////////////////////////////////////////
// 󥹥ȥ饯
////////////////////////////////////////////////////////////////
VM6::VM6( void )
{
	UDCount		= 0;		// ̹ޡ󥿽
	UDFPSCount	= 0;		// FPSޡ󥿽
	
	io		= NULL;
	sche	= NULL;
	snd		= NULL;
	psg		= NULL;
	mem		= NULL;
	graph	= NULL;
	intr	= NULL;
	cpu		= NULL;
	disk	= NULL;
	cmt		= NULL;
	cmts	= NULL;
	subcpu	= NULL;
	pio		= NULL;
	key		= NULL;
	voice   = NULL;
	
	staw    = NULL;
	
	UpDateFPSID	= NULL;
	UpDateID	= NULL;
	
	#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	bp      = NULL;
	regw	= NULL;
	memw	= NULL;
	monw	= NULL;
	#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
}

VM60::VM60( void ){}
VM62::VM62( void ){}
VM66::VM66( void ){}


////////////////////////////////////////////////////////////////
// ǥȥ饯
////////////////////////////////////////////////////////////////
VM6::~VM6( void )
{
	#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	if( monw )   delete monw;
	if( memw )   delete memw;
	if( regw )   delete regw;
	if( bp )     delete bp;
	#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	
	if( UpDateID ) SDL_RemoveTimer( UpDateID );
	if( UpDateFPSID ) SDL_RemoveTimer( UpDateFPSID );
	
	if( staw )   delete staw;
	
	if( voice )  delete voice;
	if( key )    delete key;
	if( pio )    delete pio;
	if( subcpu ) delete subcpu;
	if( cmts )   delete cmts;
	if( cmt )    delete cmt;
	if( disk )   delete disk;
	if( cpu )    delete cpu;
	if( intr )   delete intr;
	if( graph )  delete graph;
	if( mem )    delete mem;
	if( psg )    delete psg;
	if( snd )    delete snd;
	if( sche )   delete sche;
	if( io )     delete io;
}

VM60::~VM60( void ){}
VM62::~VM62( void ){}
VM66::~VM66( void ){}



////////////////////////////////////////////////////////////////
// åɴؿ(С饤)
////////////////////////////////////////////////////////////////
void VM6::OnThread( void *inst )
{
	VM6 *p6;
	
	p6 = (VM6 *)inst;	// ʬȤΥ֥ȥݥ󥿼
	
	#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	if( !bp->ExistBreakPoint() ){
	#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
		// ̾¹
		while( !this->SThread::IsCancel() )
			p6->Emu();
//			p6->EmuVSYNC();
	#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	}else{
		// ֥졼ݥȤ
		while( !this->SThread::IsCancel() ){
			p6->Emu();		// 1̿¹
			
			if( bp->CheckBreakPoint( BPoint::BP_PC, cpu->GetPC(), 0, NULL ) || bp->IsReqBreak() ){
				this->SThread::Cancel();	// åɽλե饰ΩƤ
				bp->ResetBreak();
				SENDEVENT( SDL_DEBUGMODEBP )
				break;
			}
		}
	}
	#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
}


////////////////////////////////////////////////////////////////
// Waitͭ?
////////////////////////////////////////////////////////////////
BOOL VM6::IsWaitEnable( void )
{
	return sche->GetWaitEnable() && (!cGlobal::TurboTAPE || (subcpu->GetCmtStatus() == CMTCLOSE));
}


////////////////////////////////////////////////////////////////
// 1̿¹
////////////////////////////////////////////////////////////////
int VM6::Emu( void )
{
	int state = cpu->Exec();	// CPU 1̿¹
	sche->Update( state );		// ٥ȹ(Ȥޤ)
	return state;
}


////////////////////////////////////////////////////////////////
// 1ʬ¹
////////////////////////////////////////////////////////////////
int VM6::EmuVSYNC( void )
{
	int state = 0;
	sche->WaitReset();
	while( !sche->IsVSYNC() ){
		state = cpu->Exec();		// CPU 1̿¹
		sche->Update( state );		// ٥ȹ
	}
	sche->VWait();
	
	return state;
}


#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
////////////////////////////////////////////////////////////////
// ꥹơȿ¹
////////////////////////////////////////////////////////////////
void VM6::Exec( int st )
{
	int State = st;
	
	while( State > 0 ) State -= Emu();
}


////////////////////////////////////////////////////////////////
// ˥⡼ؤ
////////////////////////////////////////////////////////////////
BOOL VM6::ToggleMonitor( void )
{
	bp->ClearStatus();
	
	// ˥ɥɽڴ
	cGlobal::MonDisp = cGlobal::MonDisp ? FALSE : TRUE;
	
	// ˥⡼ɤʤUnicodeͭ
	if( cGlobal::MonDisp ){	// ˥⡼
		SDL_EnableUNICODE( 1 );	// Unicodeͭ
		Stop();					// VM
	}else{					// ̾⡼
		SDL_EnableUNICODE( 0 );	// Unicode̵
		Start();				// VMƳ
	}
	
	// ꡼󥵡ե򥯥ꥢ
	graph->ClearScreen();
	
	return cGlobal::MonDisp;
}
#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
BOOL VM60::Init( void )
{
	// 顼å
	Error::SetError( Error::InitFailed );
	
	// Ƽ索֥ȳ
	io     = new IO6;											// I/O
	intr   = new INT60( this, DEV_ID('I', 'N', 'T', 'R') );		// 
	sche   = new SCH6( this );									// 塼
	snd    = new SND6;											// 
	cpu    = new CPU6( this, DEV_ID('C', 'P', 'U', '1') );		// CPU
	subcpu = new SUB6( this );									// SUB CPU
	mem    = new MEM60( cGlobal::UseExtRam );					// 
	graph  = new DSP60( this, DEV_ID('D', 'I', 'S', 'P') );		// 
	psg    = new PSG6( this, DEV_ID('P', 'S', 'G', '1') );		// PSG
	pio    = new PIO6( this, DEV_ID('8', '2', '5', '5') );		// 8255
	key    = new KEY60( this );									// 
	cmt    = new CMTL( this, DEV_ID('T', 'A', 'P', 'E') );		// CMT(LOAD)
	cmts   = new CMTS( this );									// CMT(SAVE)
	disk   = new DSK60( DEV_ID('D', 'I', 'S', 'K'), cGlobal::FddNum );	// DISK
	staw   = new cWndStat( this );								// ơС
	
	if( !io  || !intr || !sche || !snd || !cpu  || !subcpu || !mem || !graph ||
		!psg || !pio  || !key  || !cmt || !cmts || !disk   || !staw )
			return FALSE;
	
	
	// Ƽ索֥ȽȥǥХ³
	// I/O-----
	if( !io->Init() ) return FALSE;
	
	//  -----
	const static IOBus::Connector c_intr[] = {
		{ 0xb0, IOBus::portout, INT60::outB0H },
		{ 0, 0, 0 }	};
	if( !io->Connect( (INT60*)intr, c_intr ) ) return FALSE;
	if( !intr->Init() ) return FALSE;
	
	// 塼 -----
	if( !sche->Init( CPU_CLOCK * cGlobal::OverClock / 100 ) ) return FALSE;
	
	//  -----
	if( !snd->Init( this, VM6::SoundUpdate, cGlobal::SampleRate, cGlobal::SoundBuffer ) ) return FALSE;
	snd->SetVolume( cGlobal::MasterVol );
	
	// CPU -----
	if( !cpu->Init() ) return FALSE;
	
	// SUB CPU -----
	if( !subcpu->Init() ) return FALSE;
	
	//  -----
	if( !mem->Init( cGlobal::RomPath, cGlobal::CheckCRC ) ) return FALSE;
	if( *cGlobal::ExtRomFile ) if( !mem->MountExtRom( cGlobal::ExtRomFile ) ) return FALSE;
	
	//  -----
	const static IOBus::Connector c_graph[] = {
		{ 0xb0, IOBus::portout, DSP60::outB0H },
		{ 0, 0, 0 } };
	if( !io->Connect( (DSP60*)graph, c_graph ) ) return FALSE;
	if( !graph->Init() ) return FALSE;
	graph->SetMode4Color( cGlobal::Mode4Color );
	
	// PSG -----
	const static IOBus::Connector c_psg[] = {
		{ 0xa0, IOBus::portout, PSG6::outA0H },
		{ 0xa1, IOBus::portout, PSG6::outA1H },
		{ 0xa3, IOBus::portout, PSG6::outA3H },
		{ 0xa2, IOBus::portin,  PSG6::inA2H },
		{ 0, 0, 0 }	};
	if( !io->Connect( psg, c_psg ) ) return FALSE;
	if( !psg->Init( snd->AddStream() ) ) return FALSE;
	psg->SetVolume( cGlobal::PsgVol );
	
	// 8255 -----
	const static IOBus::Connector c_8255[] = {
		{ 0x90, IOBus::portout, PIO6::out90H },
		{ 0x91, IOBus::portout, PIO6::out91H },
		{ 0x92, IOBus::portout, PIO6::out92H },
		{ 0x93, IOBus::portout, PIO6::out93H },
		{ 0x90, IOBus::portin,  PIO6::in90H },
		{ 0x92, IOBus::portin,  PIO6::in92H },
		{ 0x93, IOBus::portin,  PIO6::in93H },
		{ 0, 0, 0 }	};
	if( !io->Connect( pio, c_8255 ) ) return FALSE;
	if( !pio->Init() ) return FALSE;
	
	//  -----
	if( !key->Init( cGlobal::KeyRepeat ) ) return FALSE;
	
	// CMT(LOAD) -----
	const static IOBus::Connector c_cmt[] = {
		{ 0xb0, IOBus::portout, CMTL::outB0H },
		{ 0, 0, 0 }	};
	if( !io->Connect( cmt, c_cmt ) ) return FALSE;
	if( !cmt->Init( snd->AddStream() ) ) return FALSE;
	cmt->SetVolume( cGlobal::CmtVol );
	// TAPE
	if( *cGlobal::TapeFile ) TapeInsert( cGlobal::TapeFile );
	
	// CMT(SAVE) -----
	if( !cmts->Init() ) return FALSE;
	
	// DISK -----
	if( disk->GetDrives() ){
		const static IOBus::Connector c_disk[] = {
			{ 0xd1, IOBus::portout, DSK60::outD1H },
			{ 0xd2, IOBus::portout, DSK60::outD2H },
			{ 0xd3, IOBus::portout, DSK60::outD3H },
			{ 0xd0, IOBus::portin,  DSK60::inD0H },
			{ 0xd1, IOBus::portin,  DSK60::inD1H },
			{ 0xd2, IOBus::portin,  DSK60::inD2H },
			{ 0xd3, IOBus::portin,  DSK60::inD3H },
			{ 0, 0, 0 }	};
		if( !io->Connect( (DSK60*)disk, c_disk ) ) return FALSE;
		if( !disk->Init() ) return FALSE;
		// ɥ饤1DISK
		if( *cGlobal::DiskFile[0] ) DiskInsert( 0, cGlobal::DiskFile[0] );
	}
	
	
	// ơС -----
	if( !staw->Init() ) return FALSE;
	
	#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	// ֥졼ݥ -----
	bp = new BPoint();
	if( !bp ) return FALSE;
	bp->ClearStatus();
	
	// 쥸ɥ -----
	regw = new cWndReg( this );
	if( !regw ) return FALSE;
	if( !regw->Init() ) return FALSE;
	
	// ꥦɥ -----
	memw = new cWndMem( this );
	if( !memw ) return FALSE;
	if( !memw->Init() ) return FALSE;
	
	// ˥ɥ -----
	monw = new cWndMon( this );
	if( !monw ) return FALSE;
	if( !monw->Init() ) return FALSE;
	#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	
	
	// ̹ޡ
	if( !SetFPS( cGlobal::FrameRate ) ){
		Stop();
		return FALSE;
	}
	
	Reset();
	
	// 顼ʤ
	Error::SetError( Error::NoError );
	
	return TRUE;
}

BOOL VM62::Init( void )
{
	// 顼å
	Error::SetError( Error::InitFailed );
	
	// Ƽ索֥ȳ
	io     = new IO6;											// I/O
	intr   = new INT62( this, DEV_ID('I', 'N', 'T', 'R') );		// 
	sche   = new SCH6( this );									// 塼
	snd    = new SND6;											// 
	cpu    = new CPU6( this, DEV_ID('C', 'P', 'U', '1') );		// CPU
	subcpu = new SUB6( this );									// SUB CPU
	mem    = new MEM62( DEV_ID('M', 'E', 'M', '1') );			// 
	graph  = new DSP62( this, DEV_ID('D', 'I', 'S', 'P') );		// 
	psg    = new PSG6( this, DEV_ID('P', 'S', 'G', '1') );		// PSG
	pio    = new PIO6( this, DEV_ID('8', '2', '5', '5') );		// 8255
	key    = new KEY6( this );									// 
	voice  = new VCE6( this, DEV_ID('V', 'O', 'I', 'C') );		// 
	cmt    = new CMTL( this, DEV_ID('T', 'A', 'P', 'E') );		// CMT(LOAD)
	cmts   = new CMTS( this );									// CMT(SAVE)
	disk   = new DSK60( DEV_ID('D', 'I', 'S', 'K'), cGlobal::FddNum );	// DISK
	staw   = new cWndStat( this );								// ơС
	
	if( !io  || !intr || !sche || !snd   || !cpu || !subcpu || !mem  || !graph ||
		!psg || !pio  || !key  || !voice || !cmt || !cmts   || !disk || !staw )
			return FALSE;
	
	
	// Ƽ索֥ȽȥǥХ³
	// I/O-----
	if( !io->Init() ) return FALSE;
	
	//  -----
	const static IOBus::Connector c_intr[] = {
		{ 0xb0, IOBus::portout, INT62::outB0H },
		{ 0xf3, IOBus::portout, INT62::outF3H },
		{ 0xf4, IOBus::portout, INT62::outF4H },
		{ 0xf5, IOBus::portout, INT62::outF5H },
		{ 0xf6, IOBus::portout, INT62::outF6H },
		{ 0xf7, IOBus::portout, INT62::outF7H },
		{ 0xf7, IOBus::portin,  INT62::inF7H },
		{ 0, 0, 0 }	};
	if( !io->Connect( (INT62*)intr, c_intr ) ) return FALSE;
	if( !intr->Init() ) return FALSE;
	
	// 塼 -----
	if( !sche->Init( CPU_CLOCK * cGlobal::OverClock / 100 ) ) return FALSE;
	
	//  -----
	if( !snd->Init( this, VM6::SoundUpdate, cGlobal::SampleRate, cGlobal::SoundBuffer ) ) return FALSE;
	snd->SetVolume( cGlobal::MasterVol );
	
	// CPU -----
	const static IOBus::Connector c_cpu[] = {
		{ 0xf3, IOBus::portout, CPU6::outF3H },
		{ 0, 0, 0 }	};
	if( !io->Connect( cpu, c_cpu ) ) return FALSE;
	if( !cpu->Init() ) return FALSE;
	
	// SUB CPU -----
	if( !subcpu->Init() ) return FALSE;
	
	//  -----
	const static IOBus::Connector c_memory[] = {
		{ 0xc1, IOBus::portout, MEM62::outC1H },
		{ 0xc2, IOBus::portout, MEM62::outC2H },
		{ 0xc3, IOBus::portout, MEM62::outC3H },
		{ 0xf0, IOBus::portout, MEM62::outF0H },
		{ 0xf1, IOBus::portout, MEM62::outF1H },
		{ 0xf2, IOBus::portout, MEM62::outF2H },
		{ 0xf0, IOBus::portin,  MEM62::inF0H },
		{ 0xf1, IOBus::portin,  MEM62::inF1H },
		{ 0xf2, IOBus::portin,  MEM62::inF2H },
		{ 0, 0, 0 }	};
	if( !io->Connect( (MEM62*)mem, c_memory ) ) return FALSE;
	if( !mem->Init( cGlobal::RomPath, cGlobal::CheckCRC ) ) return FALSE;
	if( *cGlobal::ExtRomFile ) if( !mem->MountExtRom( cGlobal::ExtRomFile ) ) return FALSE;
	
	//  -----
	const static IOBus::Connector c_graph[] = {
		{ 0xb0, IOBus::portout, DSP62::outB0H },
		{ 0xc0, IOBus::portout, DSP62::outC0H },
		{ 0xc1, IOBus::portout, DSP62::outC1H },
		{ 0xc1, IOBus::portin,  DSP62::inC1H },
		{ 0, 0, 0 } };
	if( !io->Connect( (DSP62*)graph, c_graph ) ) return FALSE;
	if( !graph->Init() ) return FALSE;
	graph->SetMode4Color( cGlobal::Mode4Color );
	
	// PSG -----
	const static IOBus::Connector c_psg[] = {
		{ 0xa0, IOBus::portout, PSG6::outA0H },
		{ 0xa1, IOBus::portout, PSG6::outA1H },
		{ 0xa3, IOBus::portout, PSG6::outA3H },
		{ 0xa2, IOBus::portin,  PSG6::inA2H },
		{ 0, 0, 0 }	};
	if( !io->Connect( psg, c_psg ) ) return FALSE;
	if( !psg->Init( snd->AddStream() ) ) return FALSE;
	psg->SetVolume( cGlobal::PsgVol );
	
	// 8255 -----
	const static IOBus::Connector c_8255[] = {
		{ 0x90, IOBus::portout, PIO6::out90H },
		{ 0x91, IOBus::portout, PIO6::out91H },
		{ 0x92, IOBus::portout, PIO6::out92H },
		{ 0x93, IOBus::portout, PIO6::out93H },
		{ 0x90, IOBus::portin,  PIO6::in90H },
		{ 0x92, IOBus::portin,  PIO6::in92H },
		{ 0x93, IOBus::portin,  PIO6::in93H },
		{ 0, 0, 0 }	};
	if( !io->Connect( pio, c_8255 ) ) return FALSE;
	if( !pio->Init() ) return FALSE;
	
	//  -----
	if( !key->Init( cGlobal::KeyRepeat ) ) return FALSE;
	
	//  -----
	const static IOBus::Connector c_voice[] = {
		{ 0xe0, IOBus::portout, VCE6::outE0H },
		{ 0xe2, IOBus::portout, VCE6::outE2H },
		{ 0xe3, IOBus::portout, VCE6::outE3H },
		{ 0xe0, IOBus::portin,  VCE6::inE0H },
		{ 0xe2, IOBus::portin,  VCE6::inE2H },
		{ 0xe3, IOBus::portin,  VCE6::inE3H },
		{ 0, 0, 0 }	};
	if( !io->Connect( voice, c_voice ) ) return FALSE;
	if( !voice->Init( snd->AddStream(), cGlobal::WavePath ) ) return FALSE;
	voice->SetVolume( cGlobal::VoiceVol );
	
	// CMT(LOAD) -----
	const static IOBus::Connector c_cmt[] = {
		{ 0xb0, IOBus::portout, CMTL::outB0H },
		{ 0, 0, 0 }	};
	if( !io->Connect( cmt, c_cmt ) ) return FALSE;
	if( !cmt->Init( snd->AddStream() ) ) return FALSE;
	cmt->SetVolume( cGlobal::CmtVol );
	// TAPE
	if( *cGlobal::TapeFile ) TapeInsert( cGlobal::TapeFile );
	
	// CMT(SAVE) -----
	if( !cmts->Init() ) return FALSE;
	
	// DISK -----
	if( disk->GetDrives() ){
		const static IOBus::Connector c_disk[] = {
			{ 0xd1, IOBus::portout, DSK60::outD1H },
			{ 0xd2, IOBus::portout, DSK60::outD2H },
			{ 0xd3, IOBus::portout, DSK60::outD3H },
			{ 0xd0, IOBus::portin,  DSK60::inD0H },
			{ 0xd1, IOBus::portin,  DSK60::inD1H },
			{ 0xd2, IOBus::portin,  DSK60::inD2H },
			{ 0xd3, IOBus::portin,  DSK60::inD3H },
			{ 0, 0, 0 }	};
		if( !io->Connect( (DSK60*)disk, c_disk ) ) return FALSE;
		if( !disk->Init() ) return FALSE;
		// ɥ饤1DISK
		if( *cGlobal::DiskFile[0] ) DiskInsert( 0, cGlobal::DiskFile[0] );
	}
	
	
	// ơС -----
	if( !staw->Init() ) return FALSE;
	
	#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	// ֥졼ݥ -----
	bp = new BPoint();
	if( !bp ) return FALSE;
	bp->ClearStatus();
	
	// 쥸ɥ -----
	regw = new cWndReg( this );
	if( !regw ) return FALSE;
	if( !regw->Init() ) return FALSE;
	
	// ꥦɥ -----
	memw = new cWndMem( this );
	if( !memw ) return FALSE;
	if( !memw->Init() ) return FALSE;
	
	// ˥ɥ -----
	monw = new cWndMon( this );
	if( !monw ) return FALSE;
	if( !monw->Init() ) return FALSE;
	#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	
	
	// ̹ޡ
	if( !SetFPS( cGlobal::FrameRate ) ){
		Stop();
		return FALSE;
	}
	
	Reset();
	
	// 顼ʤ
	Error::SetError( Error::NoError );
	
	return TRUE;
}

BOOL VM66::Init( void )
{
	// 顼å
	Error::SetError( Error::InitFailed );
	
	// Ƽ索֥ȳ
	io     = new IO6;											// I/O
	intr   = new INT62( this, DEV_ID('I', 'N', 'T', 'R') );		// 
	sche   = new SCH6( this );									// 塼
	snd    = new SND6;											// 
	cpu    = new CPU6( this, DEV_ID('C', 'P', 'U', '1') );		// CPU
	subcpu = new SUB6( this );									// SUB CPU
	mem    = new MEM66( DEV_ID('M', 'E', 'M', '1') );			// 
	graph  = new DSP62( this, DEV_ID('D', 'I', 'S', 'P') );		// 
	psg    = new PSG6( this, DEV_ID('P', 'S', 'G', '1') );		// PSG
	pio    = new PIO6( this, DEV_ID('8', '2', '5', '5') );		// 8255
	key    = new KEY6( this );									// 
	voice  = new VCE6( this, DEV_ID('V', 'O', 'I', 'C') );		// 
	cmt    = new CMTL( this, DEV_ID('T', 'A', 'P', 'E') );		// CMT(LOAD)
	cmts   = new CMTS( this );									// CMT(SAVE)
	disk   = new DSK66( DEV_ID('D', 'I', 'S', 'K'), cGlobal::FddNum );	// DISK
	staw   = new cWndStat( this );								// ơС
	
	if( !io  || !intr || !sche || !snd   || !cpu || !subcpu || !mem  || !graph ||
		!psg || !pio  || !key  || !voice || !cmt || !cmts   || !disk || !staw )
			return FALSE;
	
	
	// Ƽ索֥ȽȥǥХ³
	// I/O-----
	if( !io->Init() ) return FALSE;
	
	//  -----
	const static IOBus::Connector c_intr[] = {
		{ 0xb0, IOBus::portout, INT62::outB0H },
		{ 0xf3, IOBus::portout, INT62::outF3H },
		{ 0xf4, IOBus::portout, INT62::outF4H },
		{ 0xf5, IOBus::portout, INT62::outF5H },
		{ 0xf6, IOBus::portout, INT62::outF6H },
		{ 0xf7, IOBus::portout, INT62::outF7H },
		{ 0xf7, IOBus::portin,  INT62::inF7H },
		{ 0, 0, 0 }	};
	if( !io->Connect( (INT62*)intr, c_intr ) ) return FALSE;
	if( !intr->Init() ) return FALSE;
	
	// 塼 -----
	if( !sche->Init( CPU_CLOCK * cGlobal::OverClock / 100 ) ) return FALSE;
	
	//  -----
	if( !snd->Init( this, VM6::SoundUpdate, cGlobal::SampleRate, cGlobal::SoundBuffer ) ) return FALSE;
	snd->SetVolume( cGlobal::MasterVol );
	
	// CPU -----
	const static IOBus::Connector c_cpu[] = {
		{ 0xf3, IOBus::portout, CPU6::outF3H },
		{ 0, 0, 0 }	};
	if( !io->Connect( cpu, c_cpu ) ) return FALSE;
	if( !cpu->Init() ) return FALSE;
	
	// SUB CPU -----
	if( !subcpu->Init() ) return FALSE;
	
	//  -----
	const static IOBus::Connector c_memory[] = {
		{ 0xc1, IOBus::portout, MEM62::outC1H },
		{ 0xc2, IOBus::portout, MEM62::outC2H },
		{ 0xc3, IOBus::portout, MEM62::outC3H },
		{ 0xf0, IOBus::portout, MEM62::outF0H },
		{ 0xf1, IOBus::portout, MEM62::outF1H },
		{ 0xf2, IOBus::portout, MEM62::outF2H },
		{ 0xf0, IOBus::portin,  MEM62::inF0H },
		{ 0xf1, IOBus::portin,  MEM62::inF1H },
		{ 0xf2, IOBus::portin,  MEM62::inF2H },
		{ 0, 0, 0 }	};
	if( !io->Connect( (MEM66*)mem, c_memory ) ) return FALSE;
	if( !mem->Init( cGlobal::RomPath, cGlobal::CheckCRC ) ) return FALSE;
	if( *cGlobal::ExtRomFile ) if( !mem->MountExtRom( cGlobal::ExtRomFile ) ) return FALSE;
	
	//  -----
	const static IOBus::Connector c_graph[] = {
		{ 0xb0, IOBus::portout, DSP62::outB0H },
		{ 0xc0, IOBus::portout, DSP62::outC0H },
		{ 0xc1, IOBus::portout, DSP62::outC1H },
		{ 0xc1, IOBus::portin,  DSP62::inC1H },
		{ 0, 0, 0 } };
	if( !io->Connect( (DSP62*)graph, c_graph ) ) return FALSE;
	if( !graph->Init() ) return FALSE;
	graph->SetMode4Color( cGlobal::Mode4Color );
	
	// PSG -----
	const static IOBus::Connector c_psg[] = {
		{ 0xa0, IOBus::portout, PSG6::outA0H },
		{ 0xa1, IOBus::portout, PSG6::outA1H },
		{ 0xa3, IOBus::portout, PSG6::outA3H },
		{ 0xa2, IOBus::portin,  PSG6::inA2H },
		{ 0, 0, 0 }	};
	if( !io->Connect( psg, c_psg ) ) return FALSE;
	if( !psg->Init( snd->AddStream() ) ) return FALSE;
	psg->SetVolume( cGlobal::PsgVol );
	
	// 8255 -----
	const static IOBus::Connector c_8255[] = {
		{ 0x90, IOBus::portout, PIO6::out90H },
		{ 0x91, IOBus::portout, PIO6::out91H },
		{ 0x92, IOBus::portout, PIO6::out92H },
		{ 0x93, IOBus::portout, PIO6::out93H },
		{ 0x90, IOBus::portin,  PIO6::in90H },
		{ 0x92, IOBus::portin,  PIO6::in92H },
		{ 0x93, IOBus::portin,  PIO6::in93H },
		{ 0, 0, 0 }	};
	if( !io->Connect( pio, c_8255 ) ) return FALSE;
	if( !pio->Init() ) return FALSE;
	
	//  -----
	if( !key->Init( cGlobal::KeyRepeat ) ) return FALSE;
	
	//  -----
	const static IOBus::Connector c_voice[] = {
		{ 0xe0, IOBus::portout, VCE6::outE0H },
		{ 0xe2, IOBus::portout, VCE6::outE2H },
		{ 0xe3, IOBus::portout, VCE6::outE3H },
		{ 0xe0, IOBus::portin,  VCE6::inE0H },
		{ 0xe2, IOBus::portin,  VCE6::inE2H },
		{ 0xe3, IOBus::portin,  VCE6::inE3H },
		{ 0, 0, 0 }	};
	if( !io->Connect( voice, c_voice ) ) return FALSE;
	if( !voice->Init( snd->AddStream(), cGlobal::WavePath ) ) return FALSE;
	voice->SetVolume( cGlobal::VoiceVol );
	
	// CMT(LOAD) -----
	const static IOBus::Connector c_cmt[] = {
		{ 0xb0, IOBus::portout, CMTL::outB0H },
		{ 0, 0, 0 }	};
	if( !io->Connect( cmt, c_cmt ) ) return FALSE;
	if( !cmt->Init( snd->AddStream() ) ) return FALSE;
	cmt->SetVolume( cGlobal::CmtVol );
	// TAPE
	if( *cGlobal::TapeFile ) TapeInsert( cGlobal::TapeFile );
	
	// CMT(SAVE) -----
	if( !cmts->Init() ) return FALSE;
	
	// DISK -----
	const static IOBus::Connector c_disk[] = {
		{ 0xb1, IOBus::portout, DSK66::outB1H },
		{ 0xb2, IOBus::portout, DSK66::outB2H },
		{ 0xb3, IOBus::portout, DSK66::outB3H },
		{ 0xd0, IOBus::portout, DSK66::outD0H },
		{ 0xd1, IOBus::portout, DSK66::outD1H },
		{ 0xd2, IOBus::portout, DSK66::outD2H },
		{ 0xd3, IOBus::portout, DSK66::outD3H },
		{ 0xd6, IOBus::portout, DSK66::outD6H },
		{ 0xd8, IOBus::portout, DSK66::outD8H },
		{ 0xda, IOBus::portout, DSK66::outDAH },
		{ 0xdd, IOBus::portout, DSK66::outDDH },
		{ 0xde, IOBus::portout, DSK66::outDEH },
		{ 0xb2, IOBus::portin,  DSK66::inB2H },
		{ 0xd0, IOBus::portin,  DSK66::inD0H },
		{ 0xd1, IOBus::portin,  DSK66::inD1H },
		{ 0xd2, IOBus::portin,  DSK66::inD2H },
		{ 0xd3, IOBus::portin,  DSK66::inD3H },
		{ 0xd4, IOBus::portin,  DSK66::inD4H },
		{ 0xdc, IOBus::portin,  DSK66::inDCH },
		{ 0xdd, IOBus::portin,  DSK66::inDDH },
		{ 0, 0, 0 }	};
	if( !io->Connect( (DSK66*)disk, c_disk ) ) return FALSE;
	if( !disk->Init() ) return FALSE;
	// ɥ饤1DISK
	if( *cGlobal::DiskFile[0] ) DiskInsert( 0, cGlobal::DiskFile[0] );
	
	
	// ơС -----
	if( !staw->Init() ) return FALSE;
	
	#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	// ֥졼ݥ -----
	bp = new BPoint();
	if( !bp ) return FALSE;
	bp->ClearStatus();
	
	// 쥸ɥ -----
	regw = new cWndReg( this );
	if( !regw ) return FALSE;
	if( !regw->Init() ) return FALSE;
	
	// ꥦɥ -----
	memw = new cWndMem( this );
	if( !memw ) return FALSE;
	if( !memw->Init() ) return FALSE;
	
	// ˥ɥ -----
	monw = new cWndMon( this );
	if( !monw ) return FALSE;
	if( !monw->Init() ) return FALSE;
	#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	
	// ̹ޡ
	if( !SetFPS( cGlobal::FrameRate ) ){
		Stop();
		return FALSE;
	}
	
	Reset();
	
	// 顼ʤ
	Error::SetError( Error::NoError );
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// ư
////////////////////////////////////////////////////////////////
BOOL VM6::Start( void )
{
	// å
	if( !this->SThread::BeginThread( this ) ) return FALSE;
	this->SThread::Start();
	
	sche->Start();
	snd->Play();
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
void VM6::Stop( void )
{
	this->SThread::Cancel();	// åɽλե饰ΩƤ
	this->SThread::Waiting();	// åɽλޤԤ
	
	snd->Pause();
	sche->Stop();
}


////////////////////////////////////////////////////////////////
// ꥻå
////////////////////////////////////////////////////////////////
void VM6::Reset( void )
{
	cpu->Reset();
	intr->Reset();
	mem->Reset();
	subcpu->Reset();
	pio->Reset();
	psg->Reset();
	disk->Init();
	// ƥǥäƤ饪ȥ̵
	if( !disk->IsSystem(0) && !disk->IsSystem(1) ) cmt->Reset();
	if( voice ) voice->Reset();
}


////////////////////////////////////////////////////////////////
// TAPE 
////////////////////////////////////////////////////////////////
BOOL VM6::TapeInsert( char *filename )
{
	// TAPEޥ
	if( !cmt->Mount( filename ) ){
		cmt->Unmount();
		return FALSE;
	}
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// TAPE ӽ
////////////////////////////////////////////////////////////////
void VM6::TapeEject( void )
{
	cmt->Unmount();
}


////////////////////////////////////////////////////////////////
// DISK 
////////////////////////////////////////////////////////////////
BOOL VM6::DiskInsert( int drvno, char *filename )
{
	if( !disk->Mount( drvno, filename ) ){
		disk->Unmount( drvno );
		return FALSE;
	}
	return TRUE;
}


////////////////////////////////////////////////////////////////
// DISK ӽ
////////////////////////////////////////////////////////////////
void VM6::DiskEject( int drvno )
{
	disk->Unmount( drvno );
}


////////////////////////////////////////////////////////////////
// ĥROM 
////////////////////////////////////////////////////////////////
BOOL VM6::ExtRomInsert( char *filename )
{
	// ĥROMޥ
	if( !mem->MountExtRom( filename ) ){
		mem->UnmountExtRom();
		return FALSE;
	}
	return TRUE;
}


////////////////////////////////////////////////////////////////
// ĥROM ӽ
////////////////////////////////////////////////////////////////
void VM6::ExtRomEject( void )
{
	mem->UnmountExtRom();
}


////////////////////////////////////////////////////////////////
// ٥ȥ롼
////////////////////////////////////////////////////////////////
BOOL VM6::EventLoop( void )
{
	SDL_Event event;
	
	while( SDL_WaitEvent( &event ) ){
		switch( event.type ){
		case SDL_DRAWSCREEN:	// ̹
			#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
			if( DISPMON ){
				regw->Update();
				memw->Update();
				monw->Update();
			}
			#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
			graph->DrawScreen();
			UDCount = 0;
			break;
			
		case SDL_FPSUPDATE:		// FPS
			{
			char str[256];
			sprintf( str, "%s (%d%%)", cGlobal::WCaption, sche->GetRatio() );
			SDL_WM_SetCaption( str, "" );
			}
			UDFPSCount = 0;
			break;
			
		case SDL_KEYDOWN:		// 
			#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
			// ˥⡼?
			if( DISPMON ){
				monw->KeyIn( &event );
				break;
			}
			#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
			// Ƽﵡǽå
			if( CheckFuncKey( &event ) ) break;
		case SDL_KEYUP:
			key->ScanKeyboard( &event );	// ּ̥
			break;
			
		#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
		case SDL_MOUSEMOTION:	// ꥦɥ ɥ쥹ѹ
			if( DISPMON && event.motion.yrel ){
				// ܥ1ʺܥ)Ƥ?
				if( SDL_GetMouseState( NULL, NULL ) & SDL_BUTTON( 1 ) ){
					if( event.motion.yrel > 0 ) memw->SetAddress( memw->GetAddress() - 16 );
					else                        memw->SetAddress( memw->GetAddress() + 16 );
				}
			}
			break;
			
		case SDL_DEBUGMODEBP:	// ˥⡼ѹ(֥졼ݥã)
			// ǥ֥졼ݥȤξɽ
			
			ToggleMonitor();	// ˥⡼ѹ
			
			break;
			
		#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
		
		case SDL_MOUSEBUTTONUP:	// ޥܥ󥯥å
			if( ( event.button.button == SDL_BUTTON_RIGHT ) && ( event.button.state == SDL_RELEASED ) ){
				Stop();
				// ݥåץåץ˥塼ɽ
				ShowPopupMenu( event.button.x, event.button.y );
				#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
				if( !DISPMON )
				#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
					Start();
			}
			break;
			
		case SDL_QUIT:			// λ
			return FALSE;
			
		case SDL_RESTART:		// Ƶư
			return TRUE;
			
		}
		
		// 顼
		if( Error::GetError() != Error::NoError ){
			OSD_Message( (char *)Error::GetErrorText(), MSGE_ERROR, OSDR_OK | OSDM_ICONERROR );
			Error::SetError( Error::NoError );
		}
	}
	return FALSE;
}


////////////////////////////////////////////////////////////////
// Ƽﵡǽå
////////////////////////////////////////////////////////////////
BOOL VM6::CheckFuncKey( SDL_Event *event )
{
	// ALTξ֤
	BOOL OnALT = SDL_GetModState() & KMOD_ALT ? TRUE : FALSE;
	
	switch( event->key.keysym.sym ){	// 
	case SDLK_F6:		// ǥХå⡼ѹ or ꡼⡼ѹ
		if( OnALT )	cGlobal::FullScreen = cGlobal::FullScreen ? FALSE : TRUE;
		#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
		else        ToggleMonitor();
		#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
		break;
		
	case SDLK_F7:			// 饤⡼ѹ
		cGlobal::ScanLine = cGlobal::ScanLine ? FALSE : TRUE;
		break;
		
	case SDLK_F8:			// ⡼4顼ѹ or ơСɽѹ
		if( OnALT )
			cGlobal::StatDisp = cGlobal::StatDisp ? FALSE : TRUE;
		else{
			int c = graph->GetMode4Color();
			if( ++c > 4 ) c = 0;
			graph->SetMode4Color( c );
		}
		break;
		
//	case SDLK_F9:			// TAPE or DISK
//		if( OnALT ) SENDEVENT( SDL_DISKINSERT )
//		else        SENDEVENT( SDL_TAPEINSERT )
//		break;
		
	case SDLK_F10:			// Wait̵ͭѹ
		sche->SetWaitEnable( sche->GetWaitEnable() ? FALSE : TRUE );
		break;
		
	case SDLK_F11:			// ꥻå
		Reset();
		break;
		
	case SDLK_SYSREQ:		// ʥåץå
		graph->SnapShot( cGlobal::ImgPath );
		break;
	
	default:				// ɤǤʤ
		return FALSE;
	}
	return TRUE;
}


////////////////////////////////////////////////////////////////
// ꡼
////////////////////////////////////////////////////////////////
BYTE VM6::Read( WORD addr )
{
	BYTE data = mem->Read( addr );
	
	#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	// ֥졼ݥȥå
	if( bp->CheckBreakPoint( BPoint::BP_READ, addr, data, NULL ) ){
		PRINTD( MEM_LOG, " -> Break!" );
		bp->SetBreak();
	}
	#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	PRINTD( MEM_LOG, "\n" );
	
	return data;
}


////////////////////////////////////////////////////////////////
// 饤
////////////////////////////////////////////////////////////////
void VM6::Write( WORD addr, BYTE data )
{
	mem->Write( addr, data );
	
	#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	// ֥졼ݥȥå
	if( bp->CheckBreakPoint( BPoint::BP_WRITE, addr, data, NULL ) ){
		PRINTD( MEM_LOG, " -> Break!" );
		bp->SetBreak();
	}
	#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	PRINTD( MEM_LOG, "\n" );
}


////////////////////////////////////////////////////////////////
// INؿ
////////////////////////////////////////////////////////////////
BYTE VM6::In( int port )
{
	BYTE data = io->In( port );
	
	#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	// ֥졼ݥȥå
	if( bp->CheckBreakPoint( BPoint::BP_IN, port, data, NULL ) ){
		PRINTD( IO_LOG, " -> Break!" );
		bp->SetBreak();
	}
	#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	PRINTD( IO_LOG, "\n" );
	
	return data;
}


////////////////////////////////////////////////////////////////
// OUTؿ
////////////////////////////////////////////////////////////////
void VM6::Out( int port, BYTE data )
{
	io->Out( port, data );
	
	#if USEMONITOR	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	// ֥졼ݥȥå
	if( bp->CheckBreakPoint( BPoint::BP_OUT, port, data, NULL ) ){
		PRINTD( IO_LOG, " -> Break!" );
		bp->SetBreak();
	}
	#endif			// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ //
	PRINTD( IO_LOG, "\n" );
}


////////////////////////////////////////////////////////////////
// ɹ Хåؿ
// len:Хñ
////////////////////////////////////////////////////////////////
void VM6::SoundUpdate( void *userdata, Uint8 *stream, int len )
{
	VM6 *p6 = (VM6 *)userdata;			// ʬȤΥ֥ȥݥ󥿼
	
//	int samples = len / 2;				// ץ(16bit)
	
	p6->psg->SoundUpdate( -1 );			// PSG
//	if( p6->voice )						// 
//		p6->voice->SoundUpdate( samples );
	
	// ɥХåե
	p6->snd->Update( stream, len );
}


////////////////////////////////////////////////////////////////
// ̹ޡ Хåؿ
////////////////////////////////////////////////////////////////
Uint32 VM6::UpDate( Uint32 interval, void *obj )
{
	VM6 *p6 = (VM6 *)obj;	// ʬȤΥ֥ȥݥ󥿼
	
// åѤʤľǤ⤤롩
// ȤʤFPS?
//	p6->graph->DrawScreen();

	if( !p6->UDCount ){
		// ̹
		SENDEVENT( SDL_DRAWSCREEN )
		p6->UDCount++;
	}
	return interval;
}


////////////////////////////////////////////////////////////////
// FPSɽޡ Хåؿ
////////////////////////////////////////////////////////////////
Uint32 VM6::UpDateFPS( Uint32 interval, void *obj )
{
	VM6 *p6 = (VM6 *)obj;	// ʬȤΥ֥ȥݥ󥿼
	
	if( !p6->UDFPSCount ){
		// FPS
		SENDEVENT( SDL_FPSUPDATE )
		p6->UDFPSCount++;
	}
	return interval;
}


////////////////////////////////////////////////////////////////
// ե졼졼
////////////////////////////////////////////////////////////////
BOOL VM6::SetFPS( int fps )
{
	// ޡưʤ
	if( UpDateID )    SDL_RemoveTimer( UpDateID );
	if( UpDateFPSID ) SDL_RemoveTimer( UpDateFPSID );
	
	// ޡ
	UpDateID    = SDL_AddTimer( 1000/fps, VM6::UpDate,    this );
	UpDateFPSID = SDL_AddTimer( 1000,     VM6::UpDateFPS, this );
	
	return ( UpDateID && UpDateFPSID ) ? TRUE : FALSE;
}


////////////////////////////////////////////////////////////////
// ɤǤSAVE
////////////////////////////////////////////////////////////////
BOOL VM6::DokoDemoSave( char *filename )
{
	cIni *Ini = NULL;
	
	// Ȥꤢ顼
	Error::SetError( Error::DokoWriteFailed );
	try{
		FILE *fp = fopen( filename, "wt" );
		if( !fp ) throw Error::DokoWriteFailed;
		// ȥԤϤưöĤ
		fprintf( fp, MSDOKO_TITLE );
		fclose( fp );
		
		// ɤǤSAVEե򳫤
		Ini = new cIni();
		if( !Ini->Init( filename ) ) throw Error::DokoWriteFailed;
	}
	catch( std::bad_alloc ){	// new ˼Ԥ
		Error::SetError( Error::MemAllocFailed );
		return FALSE;
	}
	catch( Error::Errno i ){	// 㳰ȯ
		Error::SetError( i );
		if( Ini ) delete Ini;
		return FALSE;
	}
	
	cGlobal::DokoSave( Ini );
	psg->DokoSave( Ini );
	disk->DokoSave( Ini );
	graph->DokoSave( Ini );
	intr->DokoSave( Ini );
	key->DokoSave( Ini );
	subcpu->DokoSave( Ini );
	pio->DokoSave( Ini );
	sche->DokoSave( Ini );
	cpu->DokoSave( Ini );
	mem->DokoSave( Ini );
	cmt->DokoSave( Ini );
	
	Ini->Write();
	
	delete Ini;
	
	// ̵äΤǥ顼ʤ
	Error::SetError( Error::NoError );
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// ɤǤLOAD
////////////////////////////////////////////////////////////////
BOOL VM6::DokoDemoLoad( char *filename )
{
	cIni *Ini = NULL;
	
	// Ȥꤢ顼
	Error::SetError( Error::DokoReadFailed );
	try{
		// ɤǤLOADե򳫤
		Ini = new cIni();
		if( !Ini->Init( filename ) ) throw Error::DokoReadFailed;
		
		if( !cGlobal::DokoLoad( Ini ) ) throw Error::GetError();
	
	}
	catch( std::bad_alloc ){	// new ˼Ԥ
		Error::SetError( Error::MemAllocFailed );
		return FALSE;
	}
	catch( Error::Errno i ){	// 㳰ȯ
		Error::SetError( i );
		if( Ini ) delete Ini;
		return FALSE;
	}
	
	psg->DokoLoad( Ini );
	disk->DokoLoad( Ini );
	graph->DokoLoad( Ini );
	intr->DokoLoad( Ini );
	key->DokoLoad( Ini );
	subcpu->DokoLoad( Ini );
	pio->DokoLoad( Ini );
	sche->DokoLoad( Ini );
	cpu->DokoLoad( Ini );
	mem->DokoLoad( Ini );
	cmt->DokoLoad( Ini );
	
	delete Ini;
	
	// ̵äΤǥ顼ʤ
	Error::SetError( Error::NoError );
	
	return TRUE;
}
