#include <new>

#include "sound.h"
#include "error.h"


////////////////////////////////////////////////////////////////
// 󥹥ȥ饯
////////////////////////////////////////////////////////////////
cRing::cRing( void )
{
	Buffer = NULL;
	Size = Wpt = Rpt = Num = 0;
}


////////////////////////////////////////////////////////////////
// ǥȥ饯
////////////////////////////////////////////////////////////////
cRing::~cRing( void )
{
	if( Buffer ) delete[] Buffer;
}


////////////////////////////////////////////////////////////////
// 
// size:ץ
////////////////////////////////////////////////////////////////
BOOL cRing::Init( int size )
{
	PRINTD1( SND_LOG, "[cRing][Init] Size:%d\n", size );
	
	if( Buffer ) delete [] Buffer;
	
	Size = size*MULTI;	// ºݤΥХåեMULTI
	Wpt = Rpt = Num = 0;
	
	try{
		Buffer = new int[Size];
		memset( Buffer, 0, Size );
	}
	// new ˼Ԥ
	catch( std::bad_alloc ){
		Error::SetError( Error::MemAllocFailed );
		return FALSE;
	}
	
	Error::SetError( Error::NoError );
	return TRUE;
}


////////////////////////////////////////////////////////////////
// ɹ
////////////////////////////////////////////////////////////////
int cRing::Get( void )
{
	if( Num ){
		SCritical::Lock();
		int data = Buffer[Rpt++];
		if( Rpt == Size ) Rpt = 0;
		Num--;
		SCritical::UnLock();
		return data;
	}else
		return 0;
}


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
BOOL cRing::Put( int data )
{
	if( Num < Size ){
		SCritical::Lock();
		Buffer[Wpt++] = data;
		if( Wpt == Size ) Wpt = 0;
		Num++;
		SCritical::UnLock();
		return TRUE;
	}else
		return FALSE;
}


////////////////////////////////////////////////////////////////
// ̤ɥǡ
////////////////////////////////////////////////////////////////
int cRing::ReadySize( void )
{
	return min( Num, Size/MULTI );
}


////////////////////////////////////////////////////////////////
// ĤХåե
////////////////////////////////////////////////////////////////
int cRing::FreeSize( void )
{
	return Size/MULTI - min( Num, Size/MULTI );
}


////////////////////////////////////////////////////////////////
// Хåե
////////////////////////////////////////////////////////////////
int cRing::GetSize( void )
{
	return Size/MULTI;
}




////////////////////////////////////////////////////////////////
// 󥹥ȥ饯
////////////////////////////////////////////////////////////////
SndDev::SndDev( void ){
	Volume = 0;
	Buf    = NULL;
}


////////////////////////////////////////////////////////////////
// ǥȥ饯
////////////////////////////////////////////////////////////////
SndDev::~SndDev( void ){}


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
void SndDev::SetVolume( int vol )
{
	Volume = min( max( vol, 0 ), 100 );
}




////////////////////////////////////////////////////////////////
// 󥹥ȥ饯
////////////////////////////////////////////////////////////////
SND6::SND6( void ){}


////////////////////////////////////////////////////////////////
// ǥȥ饯
////////////////////////////////////////////////////////////////
SND6::~SND6( void )
{
	Pause();
	// ǥǥХĤ
	SDL_CloseAudio();
}


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
BOOL SND6::Init( void *p6, void (*callback)(void *, Uint8 *, int ), int rate, int size )
{
	PRINTD( SND_LOG, "[SND6][Init]\n" );
	
	ASpec.freq     = rate;							// ץ󥰥졼
	ASpec.format   = AUDIO_S16;						// 16ӥå椢
	ASpec.channels = 1;								// Υ
//	ASpec.samples  = ( ASpec.freq * size )/1000;	// Хåե(ץ)
	ASpec.samples  = rate * size / VSYNC_HZ;		// Хåե(ץ)
	ASpec.callback = callback;						// Хåؿλ
	ASpec.userdata = p6;							// Хåؿ˼ʬȤΥ֥ȥݥ󥿤Ϥ
	
	PRINTD1( SND_LOG, " SampleRate : %d\n",     ASpec.freq );
	PRINTD1( SND_LOG, " Format     : 0x%04X\n", ASpec.format );
	PRINTD1( SND_LOG, " Channels   : %d\n",     ASpec.channels );
	PRINTD1( SND_LOG, " Samples    : %d\n",     ASpec.samples );
	
	for( int i=0; i<MAXSTREAM; i++ ) RB[i] = NULL;
	
	SampleRate = rate;
	Volume     = DEFAULT_MASTERVOL;
	
	// ǥǥХ򳫤
	if( SDL_OpenAudio( &ASpec, NULL ) < 0 ) return FALSE;
	
	return TRUE;
}


////////////////////////////////////////////////////////////////
// ɹؿ
// len:Хñ
////////////////////////////////////////////////////////////////
void SND6::Update( Uint8 *stream, int len )
{
	#if SND_LOG
		fprintf( stdout, "[SND6][Update] Stream:%08X Len:%d\n", (int)stream, len/sizeof(Sint16) );
		{ for( int i=0; i<MAXSTREAM; i++ ) if( RB[i] ) fprintf( stdout, "  [%d] Len:%d\n", i, RB[i]->ReadySize() ); }
	#endif
	
	Sint16 *str = (Sint16 *)stream;
	int samples = len/sizeof(Sint16);
	
	while( samples-- ){
		int dat = 0;
		for( int i=0; i<MAXSTREAM; i++ ){
			if( RB[i] )	dat += RB[i]->Get();
		}
		dat = ( dat * Volume ) / 100;
		dat = min( max( dat, -32768 ), 32767 );
		*(str++) = (Sint16)dat;
	}
	#if SND_LOG
		{ for( int i=0; i<MAXSTREAM; i++ ) if( RB[i] ) fprintf( stdout, "  [%d] Len:%d\n", i, RB[i]->ReadySize() ); }
	#endif
}


////////////////////////////////////////////////////////////////
// ȥ꡼ɲ
////////////////////////////////////////////////////////////////
cRing *SND6::AddStream( void )
{
	PRINTD( SND_LOG, "[SND6][AddStream]\n" );
	
	int i;
	
	for( i=0; i<MAXSTREAM; i++ ){
		if( !RB[i] ){
			try{
				RB[i] = new cRing;
			}
			// new ˼Ԥ
			catch( std::bad_alloc ){
				Error::SetError( Error::MemAllocFailed );
				return FALSE;
			}
			Error::SetError( Error::NoError );
			
			// Хåեϥץ(ºݤΥХåեMULTIܤ)
			if( !RB[i]->Init( ASpec.samples ) ){
				delete RB[i];
				RB[i] = NULL;
				return RB[i];
			}
			break;
		}
	}
	
	if( i == MAXSTREAM ) return NULL;
	else                 return RB[i];
}


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
void SND6::Play( void )
{
	PRINTD( SND_LOG, "[SND6][Play]\n" );
	
	SDL_PauseAudio( 0 );
}


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
void SND6::Pause( void )
{
	PRINTD( SND_LOG, "[SND6][Pause]\n" );
	
	SDL_PauseAudio( 1 );
}


////////////////////////////////////////////////////////////////
// ޥ
////////////////////////////////////////////////////////////////
void SND6::SetVolume( int vol )
{
	Volume = min( max( vol, 0 ), 100 );
}
